from typing import Union

import hydra
import pytorch_lightning
import torch
from hydra.utils import instantiate
from omegaconf import DictConfig
from pytorch_lightning import Trainer

from kazu.distillation.models import (
    SequenceTaggingDistillationForFinalLayer,
    SequenceTaggingDistillationForIntermediateLayer,
)
from kazu.utils.constants import HYDRA_VERSION_BASE


@hydra.main(version_base=HYDRA_VERSION_BASE, config_path="../../conf", config_name="config")
def start(cfg: DictConfig) -> None:

    if torch.backends.cudnn.is_available():  # for reproducibility
        torch.backends.cudnn.deterministic = cfg.DistillationTraining.cudnn.deterministic
        torch.backends.cudnn.benchmark = (
            cfg.DistillationTraining.cudnn.benchmark  # set true if the model is not for research; True will make training faster
        )

    pytorch_lightning.seed_everything(cfg.DistillationTraining.seed)
    trainer: Trainer = instantiate(cfg.DistillationTraining.trainer)
    model: Union[
        SequenceTaggingDistillationForFinalLayer, SequenceTaggingDistillationForIntermediateLayer
    ] = instantiate(cfg.DistillationTraining.model)
    trainer.fit(model)


if __name__ == "__main__":
    start()



from functools import wraps
import traceback
from typing import Any, Callable, List, Protocol, Tuple, TypeVar, Iterable

from kazu.data.data import Document, PROCESSING_EXCEPTION
from kazu.ontology_preprocessing.base import OntologyParser


class Step(Protocol):
    @classmethod
    def namespace(cls) -> str:
        """Metadata to name/describe the step, used in various places.

        Defaults to  ``cls.__name__``.
        """
        # Note: it would be nice for this to be a property, but you can't use @classmethod
        # and @property together in most versions of python (you could in 3.9 and 3.10), and
        # it being a classmethod is quite useful.
        return cls.__name__

    def __call__(self, docs: List[Document]) -> Tuple[List[Document], List[Document]]:
        """Process documents and respond with processed and failed documents.

        Note that many steps will be decorated by :func:`~.kazu.steps.step.document_iterating_step`
        or :func:`~.kazu.steps.step.document_batch_step` which will modify the 'original'
        ``__call__`` function signature to match the expected signature for a step, as the
        decorators handle the exception/failed documents logic for you.

        :param docs:
        :return: The first element is all the provided docs (now modified by the processing), the
            second is the docs that failed to (fully) process correctly.
        """
        raise NotImplementedError


class ParserDependentStep(Step):
    """A step that depends on ontology parsers in any form.

    Steps that need information from parsers should subclass this class, in order
    for the internal databases to be correctly populated. Generally, these will be
    steps that have anything to do with Entity Linking.
    """

    def __init__(self, parsers: Iterable[OntologyParser]):
        """

        :param parsers: parsers that this step requires
        """
        for parser in parsers:
            parser.populate_databases(force=False)


Self = TypeVar("Self")
"""A TypeVar for the type of the class whose method is decorated with
:func:`~.kazu.steps.step.document_iterating_step` or
:func:`~.kazu.steps.step.document_batch_step`\\ ."""


def document_iterating_step(
    per_doc_callable: Callable[[Self, Document], Any]
) -> Callable[[Self, List[Document]], Tuple[List[Document], List[Document]]]:
    """Handle a list of :class:`~kazu.data.data.Document`\\ s and add error handling.

    Use this to decorate a method that processes a single :class:`~kazu.data.data.Document`\\ .
    The resulting method will then iterate over a list of
    :class:`~kazu.data.data.Document`\\ s, calling the decorated function for each
    :class:`~kazu.data.data.Document`\\ . Errors are handled automatically and added to the
    ``PROCESSING_EXCEPTION`` metadata of documents, with failed docs returned as the second element
    of the return value, as expected by :meth:`Step.__call__`\\ .

    Generally speaking, it will save effort and repetition to decorate a :class:`Step` with either
    :func:`~.kazu.steps.step.document_iterating_step` or
    :func:`~.kazu.steps.step.document_batch_step`\\, rather than implementing the error handling in
    the :class:`Step` itself.

    Normally, :func:`~.kazu.steps.step.document_iterating_step` would be used in preference to
    :func:`~.kazu.steps.step.document_batch_step`\\, unless the method involves computation which
    is more efficient when run in a batch, such as inference with a transformer-based Machine
    Learning model, or using spacy's `pipe <https://spacy.io/api/language/#pipe>`_ method.

    Note that this will only work for a method of a class, rather than a standalone function,
    as it expects to have to pass through 'self' as a parameter.

    :param per_doc_callable: A function that processes a single document, that you want to use as
        the :literal:`__call__` method of a :class:`Step`\\ . This must do its work by mutating the
        input document: the return value is ignored.
    """
    # note - this excludes __annotations__ as the returned function has a different type signature
    @wraps(per_doc_callable, assigned=("__module__", "__name__", "__qualname__", "__doc__"))
    def step_call(self: Self, docs: List[Document]) -> Tuple[List[Document], List[Document]]:
        failed_docs = []
        for doc in docs:
            try:
                per_doc_callable(self, doc)
            except Exception:
                doc.metadata[PROCESSING_EXCEPTION] = traceback.format_exc()
                failed_docs.append(doc)
        return docs, failed_docs

    return step_call


def document_batch_step(
    batch_doc_callable: Callable[[Self, List[Document]], Any]
) -> Callable[[Self, List[Document]], Tuple[List[Document], List[Document]]]:
    """Add error handling to a method that processes batches of :class:`~kazu.data.data.Document`\\ s.

    Use this to decorate a method that processes a batch of :class:`~kazu.data.data.Document`\\ s
    at a time. The resulting method will wrap a call to the decorated function with error handling
    which will add exceptions to the ``PROCESSING_EXCEPTION`` metadata of documents. Failed
    documents will be returned as the second element of the return value, as expected by
    :meth:`Step.__call__`\\ .

    Generally speaking, it will save effort and repetition to decorate a :class:`Step` with either
    :func:`~.kazu.steps.step.document_iterating_step` or
    :func:`~.kazu.steps.step.document_batch_step`\\, rather than implementing the error handling in
    the :class:`Step` itself.

    Normally, :func:`~.kazu.steps.step.document_iterating_step` would be used in preference to
    :func:`~.kazu.steps.step.document_batch_step`\\, unless the method involves computation which
    is more efficient when run in a batch, such as inference with a transformer-based Machine
    Learning model, or using spacy's `pipe <https://spacy.io/api/language/#pipe>`_ method.

    Note that this will only work for a method of a class, rather than a standalone function,
    as it expects to have to pass through 'self' as a parameter.

    :param batch_doc_callable: A function that processes a batch of documents, that you want to use
        as the :literal:`__call__` method of a :class:`Step`. This must do its work by mutating the
        input documents: the return value is ignored.
    """
    # note - this excludes __annotations__ as the returned function has a different type signature
    @wraps(batch_doc_callable, assigned=("__module__", "__name__", "__qualname__", "__doc__"))
    def step_call(self: Self, docs: List[Document]) -> Tuple[List[Document], List[Document]]:
        failed_docs = []
        try:
            batch_doc_callable(self, docs)
        except Exception:
            affected_doc_ids = [doc.idx for doc in docs]
            for doc in docs:
                message = (
                    f"batch failed: affected ids: {affected_doc_ids}\n" + traceback.format_exc()
                )
                doc.metadata[PROCESSING_EXCEPTION] = message
                failed_docs.append(doc)
        return docs, failed_docs

    return step_call



import logging
from pathlib import Path
from typing import List, Dict, Tuple, Union, Iterable, Sequence, overload, Type, Any

from transformers import BatchEncoding, PreTrainedTokenizerBase
from transformers.file_utils import PaddingStrategy
from transformers.tokenization_utils_base import TruncationStrategy

from kazu.data.data import Document, Entity, Section

logger = logging.getLogger(__name__)


def find_document_from_entity(docs: List[Document], entity: Entity) -> Document:
    """
    for a given entity and a list of docs, find the doc the entity belongs to

    :param list_map:
    :param entity:
    :return:
    """
    for doc in docs:
        if entity in doc.get_entities():
            return doc
    raise RuntimeError(f"Error! Entity {entity}is not attached to a document")


def documents_to_document_section_text_map(docs: List[Document]) -> Dict[Tuple[int, int], str]:
    """
    convert documents into a dict of <dochash + sectionhash>: text

    :param docs:
    :return:
    """
    return {
        (
            hash(doc),
            hash(section),
        ): section.get_text()
        for doc in docs
        for section in doc.sections
    }


def documents_to_id_section_map(docs: List[Document]) -> Dict[int, Section]:
    """
    return a map of documents, indexed by order of sections

    :param docs:
    :return:
    """
    result = {}
    i = 0
    for doc in docs:
        for section in doc.sections:
            result[i] = section
            i += 1
    return result


def filter_entities_with_ontology_mappings(entities: List[Entity]) -> List[Entity]:
    """
    finds entities that have no kb mappings

    :param entities:
    :return:
    """
    return [x for x in entities if len(x.mappings) == 0]


def documents_to_document_section_batch_encodings_map(
    docs: List[Document],
    tokenizer: PreTrainedTokenizerBase,
    stride: int = 128,
    max_length: int = 512,
) -> Tuple[BatchEncoding, Dict[int, Section]]:
    """
    convert documents into a BatchEncoding. Also returns a list of <int + section> for the resulting encoding

    :param docs:
    :return:
    """
    id_section_map = documents_to_id_section_map(docs)
    strings = [section.get_text() for section in id_section_map.values()]

    batch_encodings = tokenizer(
        strings,
        stride=stride,
        max_length=max_length,
        truncation=TruncationStrategy.LONGEST_FIRST,
        return_overflowing_tokens=True,
        padding=PaddingStrategy.MAX_LENGTH,
    )
    return batch_encodings, id_section_map


def get_match_entity_class_hash(ent: Entity) -> int:
    return hash(
        (
            ent.match,
            ent.entity_class,
        )
    )


PathLike = Union[str, Path]

SinglePathLikeOrIterable = Union[PathLike, Iterable[PathLike]]


def as_path(p: PathLike) -> Path:
    return p if isinstance(p, Path) else Path(p)


def get_cache_dir(path: PathLike, prefix: str = "", create_if_not_exist: bool = True) -> Path:
    path = as_path(path)
    new_path = path.with_name(f"cached_{prefix}_{path.name}")
    if create_if_not_exist:
        if new_path.exists():
            logger.info(f"{new_path} already exists. Will not make it")
        else:
            new_path.mkdir()

    return new_path


def get_cache_path(path: PathLike, cache_id: str) -> Path:
    path = as_path(path)
    original_filename = path.name
    cache_dir = get_cache_dir(path, create_if_not_exist=False)
    new_path = cache_dir.joinpath(f"cached_{cache_id}_{original_filename}")
    return new_path


class EntityClassFilter:
    """
    A condition that returns True if a document has any entities that match the class of the required_entity_classes
    """

    def __init__(self, required_entity_classes: Iterable[str]):
        """

        :param required_entity_classes: list of str, specifying entity classes to assess
        """
        self.required_entities = set(required_entity_classes)

    def __call__(self, document: Document) -> bool:
        return any(
            (entity.entity_class in self.required_entities for entity in document.get_entities())
        )


@overload
def _create_ngrams_iter(tokens: str, n: int) -> Iterable[str]:
    pass


@overload
def _create_ngrams_iter(tokens: Sequence[str], n: int) -> Iterable[Sequence[str]]:
    pass


def _create_ngrams_iter(tokens, n=2):
    """Yields ngrams of the input as a sequence of strings.

    Tokens can be a single string where each token is a character, or an Iterable of words.
    If tokens is a str"""
    num_tokens = len(tokens)
    for i in range(num_tokens):
        ngram_end_index = i + n
        if ngram_end_index > num_tokens:
            # ngram would extend beyond end of parts
            # it's ok for it to be the same number though as otherwise
            # we don't get the final word of ngrams at the right of parts
            break
        yield tokens[i : i + n]


def create_char_ngrams(s: str, n: int = 2) -> List[str]:
    """Return list of char ngrams as a string."""
    return list(_create_ngrams_iter(s, n))


def create_word_ngrams(s: str, n: int = 2) -> List[str]:
    """Return list of word ngrams as a single space-separated string."""
    words = s.split(" ")
    ngrams_iter = _create_ngrams_iter(words, n)
    return [" ".join(ngram) for ngram in ngrams_iter]


class Singleton(type):
    _instances: Dict[Type, Any] = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]

    @staticmethod
    def clear_all():
        logger.warning(
            "When clearing singletons, check that instances of classes with metaclass=Singleton are "
            "not used as class fields, as this will cause unexpected behaviour. Also note that any existing "
            "classes that use a Singleton will need to be re-instantiated to work correctly!. Much pain "
            "can be avoided by observing this simple principle! Some people consider Singletons an "
            "anti-pattern, and this is a reason why!!!"
        )
        Singleton._instances.clear()



from typing import List

import stanza
from stanza.models.common.doc import Sentence

from kazu.data.data import Document, CharSpan
from kazu.steps import Step, document_iterating_step


class StanzaStep(Step):
    """
    Currently just provides sentence-segmentation using a tokenizer trained on the genia treebank.

    Stanza paper:

    Peng Qi, Yuhao Zhang, Yuhui Zhang, Jason Bolton and Christopher D. Manning. 2020.
    `Stanza: A Python Natural Language Processing Toolkit for Many Human Languages. <https://arxiv.org/abs/2003.07082>`_
    In Association for Computational Linguistics (ACL) System Demonstrations. 2020.
    [`pdf <https://nlp.stanford.edu/pubs/qi2020stanza.pdf>`_][`bib <https://nlp.stanford.edu/pubs/qi2020stanza.bib>`_]

    Stanza biomedical and clinical models:

    Yuhao Zhang, Yuhui Zhang, Peng Qi, Christopher D. Manning, Curtis P. Langlotz.
    `Biomedical and Clinical English Model Packages in the Stanza Python NLP Library <https://doi.org/10.1093/jamia/ocab090>`_,
    Journal of the American Medical Informatics Association. 2021.

    .. raw:: html

        <details>
        <summary>Bibtex Citation Details (both papers above)</summary>

    .. code:: bibtex

        @inproceedings{qi2020stanza,
            author = {Qi, Peng and Zhang, Yuhao and Zhang, Yuhui and Bolton, Jason and Manning, Christopher D.},
            booktitle = {Proceedings of the 58th Annual Meeting of the Association for Computational Linguistics: System
            Demonstrations},
            title = {Stanza: A {Python} Natural Language Processing Toolkit for Many Human Languages},
            url = {https://nlp.stanford.edu/pubs/qi2020stanza.pdf},
            year = {2020}
        }

        @Article{10.1093/jamia/ocab090,
            author = {Zhang, Yuhao and Zhang, Yuhui and Qi, Peng and Manning, Christopher D and Langlotz, Curtis P},
            title = "{Biomedical and clinical English model packages for the Stanza Python NLP library}",
            journal = {Journal of the American Medical Informatics Association},
            volume = {28},
            number = {9},
            pages = {1892-1899},
            year = {2021},
            month = {06},
            abstract = "{The study sought to develop and evaluate neural natural language processing (NLP) packages for the
            syntactic analysis and named entity recognition of biomedical and clinical English text.We implement and train
            biomedical and clinical English NLP pipelines by extending the widely used Stanza library originally designed
            for general NLP tasks. Our models are trained with a mix of public datasets such as the CRAFT treebank as well
            as with a private corpus of radiology reports annotated with 5 radiology-domain entities. The resulting
            pipelines are fully based on neural networks, and are able to perform tokenization, part-of-speech tagging,
            lemmatization, dependency parsing, and named entity recognition for both biomedical and clinical text. We
            compare our systems against popular open-source NLP libraries such as CoreNLP and scispaCy, state-of-the-art
            models such as the BioBERT models, and winning systems from the BioNLP CRAFT shared task.For syntactic
            analysis, our systems achieve much better performance compared with the released scispaCy models and CoreNLP
            models retrained on the same treebanks, and are on par with the winning system from the CRAFT shared task. For
            NER, our systems substantially outperform scispaCy, and are better or on par with the state-of-the-art
            performance from BioBERT, while being much more computationally efficient.We introduce biomedical and clinical
            NLP packages built for the Stanza library. These packages offer performance that is similar to the state of the
            art, and are also optimized for ease of use. To facilitate research, we make all our models publicly available.
            We also provide an online demonstration (http://stanza.run/bio).}",
            issn = {1527-974X},
            doi = {10.1093/jamia/ocab090},
            url = {https://doi.org/10.1093/jamia/ocab090},
            eprint = {https://academic.oup.com/jamia/article-pdf/28/9/1892/39731803/ocab090.pdf},
        }

    .. raw:: html

        </details>

    """

    def __init__(self, stanza_pipeline: stanza.Pipeline):
        """

        :param stanza_pipeline: The stanza pipeline the step uses for sentence-segmentation
        """
        self.stanza_pipeline = stanza_pipeline

    @document_iterating_step
    def __call__(self, doc: Document) -> None:
        for section in doc.sections:
            stanza_doc = self.stanza_pipeline(section.get_text())
            sentences: List[Sentence] = stanza_doc.sentences
            char_spans = (
                CharSpan(sent.tokens[0].start_char, sent.tokens[-1].end_char) for sent in sentences
            )
            section.sentence_spans = char_spans



from collections import defaultdict

from kazu.data.data import Document, Entity
from kazu.steps import Step, document_iterating_step

import spacy


class SpacyNerStep(Step):
    """
    A simple spacy NER implementation. Runs a spacy pipeline over document sections, expecting the
    resulting spacy doc to have a populated doc.ents field.
    """

    def __init__(self, spacy_pipeline: spacy.Language):
        """

        :param model_name: name of spacy pipeline to load.
        :param path: If the spacy pipeline is not already installed into the python environment, attempt to
            install it from this path.
        """
        self.nlp = spacy_pipeline

    @document_iterating_step
    def __call__(self, doc: Document) -> None:
        for section in doc.sections:
            spacy_doc = self.nlp(section.get_text())
            for ent in spacy_doc.ents:
                section.entities.append(
                    Entity.load_contiguous_entity(
                        start=ent.start_char,
                        end=ent.end_char,
                        match=section.get_text()[ent.start_char : ent.end_char],
                        entity_class=ent.label_.lower(),
                        namespace=self.namespace(),
                    )
                )

            sent_metadata = defaultdict(list)
            for sent in spacy_doc.sents:
                sent_metadata["scispacy_sent_offsets"].append([sent.start_char, sent.end_char])
            if not section.metadata:
                section.metadata = {}
            section.metadata.update(sent_metadata)



import logging
from copy import deepcopy
from itertools import groupby
from typing import List, Tuple, Iterable, Dict, FrozenSet, Set, Optional

from kazu.data.data import (
    Document,
    Entity,
    SynonymTermWithMetrics,
    Mapping,
    MentionConfidence,
)
from kazu.steps.linking.post_processing.mapping_strategies.strategies import MappingStrategy
from kazu.steps.linking.post_processing.xref_manager import CrossReferenceManager
from kazu.utils.grouping import sort_then_group
from kazu.utils.string_normalizer import StringNormalizer

logger = logging.getLogger(__name__)


EntityClassStrategy = Dict[str, List[MappingStrategy]]


# hashable representation of all identical (for the purposes of mapping) entities in the document
EntityKey = Tuple[str, str, str, FrozenSet[SynonymTermWithMetrics]]


def entity_to_entity_key(
    e: Entity,
) -> EntityKey:
    return (
        e.match,
        e.match_norm,
        e.entity_class,
        frozenset(e.syn_term_to_synonym_terms),
    )


class ConfidenceLevelStrategyExecution:
    """
    The role of this class is to track which entities have had mappings successfully resolved,
    and which require the application of further strategies.

    This is handled via tracking a dictionary of EntityKey to sets of parser names.

    See further details in the __call__ docstring.
    """

    def __init__(
        self,
        ent_class_strategies: EntityClassStrategy,
        default_strategies: List[MappingStrategy],
        stop_on_success: bool = False,
    ):
        """

        :param ent_class_strategies: per class strategies
        :param default_strategies: default strategies
        """
        self.stop_on_success = stop_on_success
        self.default_strategies = default_strategies
        self.ent_class_strategies = ent_class_strategies
        self.unresolved_parsers: Dict[EntityKey, Set[str]] = {}
        self.entity_mapped: Dict[EntityKey, bool] = {}

    @property
    def longest_mapping_strategy_list_size(self):
        return max(
            (
                len(self.default_strategies),
                *(len(strategies) for strategies in self.ent_class_strategies.values()),
            )
        )

    def get_strategies_for_entity_class(self, entity_class: str) -> List[MappingStrategy]:
        return self.ent_class_strategies.get(entity_class, self.default_strategies)

    def _get_unresolved_parsers(self, entity_key: EntityKey, entity: Entity) -> Set[str]:

        maybe_unresolved_parsers = self.unresolved_parsers.get(entity_key, None)
        if maybe_unresolved_parsers is not None:
            return maybe_unresolved_parsers
        else:
            unresolved_parsers = set(x.parser_name for x in entity.syn_term_to_synonym_terms)
            self.unresolved_parsers[entity_key] = unresolved_parsers
            return unresolved_parsers

    def __call__(
        self, entity: Entity, strategy_index: int, document: Document
    ) -> Iterable[Mapping]:
        """

        Conditionally execute a mapping strategy over an entity.

        :param entity: entity to process
        :param strategy_index: index of strategy to run that is configured for this entity class
        :param document: originating Document
        :return:
        """
        strategy_list: List[MappingStrategy] = self.get_strategies_for_entity_class(
            entity_class=entity.entity_class
        )
        if strategy_index > len(strategy_list) - 1:
            logger.debug("no more strategies this class")
        else:
            strategy = strategy_list[strategy_index]
            entity_key = entity_to_entity_key(entity)
            # we keep track of which entities have resolved mappings to specific parsers, so we don't run lower
            # ranked strategies if we don't need to
            unresolved_parsers = self._get_unresolved_parsers(entity_key, entity)
            if len(unresolved_parsers) == 0:
                logger.debug(
                    f"will not run strategy {strategy.__class__.__name__} on class :<{entity.entity_class}>, match: "
                    f"<{entity.match}> as all parsers have been resolved"
                )
            elif self.stop_on_success and self.entity_mapped.get(entity_key, False):
                logger.debug(
                    f"will not run strategy {strategy.__class__.__name__} on class :<{entity.entity_class}>, match: "
                    f"<{entity.match}> as entity has been resolved to another parser and stop_on_success: "
                    f"{self.stop_on_success}"
                )
            else:
                logger.debug(
                    f"running strategy {strategy.__class__.__name__} on class :<{entity.entity_class}>, match: "
                    f"<{entity.match}> "
                )
                strategy.prepare(document)
                terms_to_consider = (
                    t
                    for t in entity.syn_term_to_synonym_terms
                    if t.parser_name in unresolved_parsers
                )
                terms_by_parser = sort_then_group(
                    terms_to_consider, key_func=lambda x: x.parser_name
                )

                for parser_name, terms_this_parser in terms_by_parser:
                    terms_this_parser_set = frozenset(terms_this_parser)
                    for mapping in strategy(
                        ent_match=entity.match,
                        ent_match_norm=entity.match_norm,
                        terms=terms_this_parser_set,
                        document=document,
                    ):
                        self.unresolved_parsers[entity_key].discard(mapping.parser_name)
                        self.entity_mapped[entity_key] = True
                        yield mapping

    def reset(self):
        """
        clear state, ready for another execution.

        Should be called when the underlying :class:`.Document` has changed.
        """
        self.unresolved_parsers.clear()
        self.entity_mapped.clear()


CONFIDENCE_ERROR_ENDING = "If entities with this confidence(s) appear, attempting to process them would cause an exception. It is likely that this class is mis-configured."


class StrategyRunner:
    """
    This is a complex class, designed to co-ordinate the running of various strategies over a document, with the end
    result producing mappings (grounding) for entities. Strategies that produce mappings may depend on the changing
    state of the Document, depending on whether other strategies are successful or not, hence why their precise
    co-ordination is crucial. Specifically we want the strategies that have higher precision to run before lower
    precision ones.

    Beyond the precision of the strategy itself, the variables to consider are:

    1. the confidence of the NER systems in the match, in that different systems vary in terms of precision and recall for detecting
       entity spans.
    2. what SynonymTerms are associated with the entity, and from which parser they originated from.

    The __call__ method of this class operates as follows:

    1. group entities by order of :class:`.MentionConfidence`\\.
    2. sub-group these entities again by :attr:`.Entity.match` and
       :attr:`.Entity.entity_class`\\ .
    3. divide these entities by whether they are symbolic or not.
    4. identify the maximum number of strategies that 'could' run.
    5. get the appropriate :class:`ConfidenceLevelStrategyExecution` to run against this sub group.
    6. group the entities from 5. by EntityKey (i.e. a hashable representation of unique information required for
       mapping.
    7. conditionally execute the next strategy out of the maximum possible (from 4), and attach any resulting mappings
       to the relevant entity group. Note, the :class:`ConfidenceLevelStrategyExecution` is responsible for deciding whether
       a strategy is executed or not.
    """

    def __init__(
        self,
        symbolic_strategies: Dict[str, ConfidenceLevelStrategyExecution],
        non_symbolic_strategies: Dict[str, ConfidenceLevelStrategyExecution],
        cross_ref_managers: Optional[List[CrossReferenceManager]] = None,
    ):
        """


        :param symbolic_strategies: mapping of mention confidence to a :class:`ConfidenceLevelStrategyExecution` for symbolic
            entities
        :param non_symbolic_strategies: mapping of mention confidence to a :class:`ConfidenceLevelStrategyExecution` for
            non-symbolic entities
        :param cross_ref_managers: list of managers that will be applied to any created mappings, attempting to create
            xreferences
        """
        self.symbolic_strategies = {MentionConfidence[k]: v for k, v in symbolic_strategies.items()}
        self.non_symbolic_strategies = {
            MentionConfidence[k]: v for k, v in non_symbolic_strategies.items()
        }
        self.cross_ref_managers = cross_ref_managers

    @staticmethod
    def group_entities_by_symbolism(
        entities: Iterable[Entity],
    ) -> Tuple[List[Entity], List[Entity]]:
        """
        groups entities into symbolic and non-symbolic forms, so they can be processed separately.

        :param entities:
        :return:
        """
        symbolic: List[Entity] = []
        non_symbolic: List[Entity] = []
        grouped_by_match = sort_then_group(
            entities,
            key_func=lambda x: (
                x.match,
                x.entity_class,
            ),
        )
        for (match_str, entity_class), ent_iter in grouped_by_match:
            if StringNormalizer.classify_symbolic(match_str, entity_class=entity_class):
                symbolic.extend(ent_iter)
            else:
                non_symbolic.extend(ent_iter)
        return symbolic, non_symbolic

    def __call__(self, doc: Document):
        """

        Run relevant strategies to decide what mappings to create.

        Generally speaking, noun phrases should be easier to normalise than symbolic mentions, as there is more
        information to work with. Therefore, we group entities by mention confidence, split by symbolism, then
        run :meth:`execute_hit_post_processing_strategies`\\ .

        :param doc:
        :return:
        """

        # do a separate sorted and groupby call (rather than our sort_then_group utility)
        # so we can do all the sorting we need in one go
        # we inverse the sign of mention_confidence so that we process high confidence
        # hits first
        sorted_entities = sorted(
            doc.get_entities(),
            key=lambda ent: (
                -ent.mention_confidence,
                *entity_to_entity_key(ent),
            ),
        )
        entities_grouped_by_confidence = groupby(
            sorted_entities, key=lambda ent: ent.mention_confidence
        )

        for mention_confidence, entities in entities_grouped_by_confidence:
            logger.debug("mapping entities for confidence %s", mention_confidence)
            symbolic_entities, non_symbolic_entities = self.group_entities_by_symbolism(
                entities=entities
            )

            maybe_non_symbolic_strategies = self.non_symbolic_strategies.get(mention_confidence)
            if maybe_non_symbolic_strategies is not None:
                self.execute_hit_post_processing_strategies(
                    non_symbolic_entities, doc, maybe_non_symbolic_strategies
                )
            else:
                logger.warning(
                    "No %s configured for %s ",
                    ConfidenceLevelStrategyExecution.__name__,
                    mention_confidence,
                )
            maybe_symbolic_strategies = self.symbolic_strategies.get(mention_confidence)
            if maybe_symbolic_strategies is not None:
                self.execute_hit_post_processing_strategies(
                    symbolic_entities, doc, maybe_symbolic_strategies
                )
            else:
                logger.warning(
                    "No %s configured for %s ",
                    ConfidenceLevelStrategyExecution.__name__,
                    mention_confidence,
                )

    def execute_hit_post_processing_strategies(
        self,
        ents_needing_mappings: List[Entity],
        document: Document,
        confidence_strategy_execution: ConfidenceLevelStrategyExecution,
    ):
        """
        This method executes parts 5 - 7 in the class Docstring.

        :param ents_needing_mappings: Expects entities to already be sorted based on :func:`entity_to_entity_key`\\ .
        :param document:
        :param confidence_strategy_execution:
        :return:
        """
        try:
            strategy_max_index = confidence_strategy_execution.longest_mapping_strategy_list_size

            groups = [
                list(ents)
                for _entity_key, ents in groupby(ents_needing_mappings, key=entity_to_entity_key)
            ]

            for i in range(0, strategy_max_index):
                for entity_group in groups:
                    reference_entity = next(iter(entity_group))
                    for mapping in confidence_strategy_execution(
                        entity=reference_entity,
                        strategy_index=i,
                        document=document,
                    ):
                        xref_mappings: Set[Mapping] = set()
                        if self.cross_ref_managers is not None:
                            for xref_manager in self.cross_ref_managers:
                                xref_mappings.update(
                                    xref_manager.create_xref_mappings(mapping=mapping)
                                )

                        for entity in entity_group:
                            entity.mappings.add(deepcopy(mapping))
                            entity.mappings.update(deepcopy(xref_mappings))
                        logger.debug(
                            "mapping created: original string: %s, mapping: %s, cross-references: %s",
                            reference_entity.match,
                            mapping,
                            xref_mappings,
                        )
        finally:
            # in case exception is thrown - always reset state
            confidence_strategy_execution.reset()



import logging
from typing import List, Tuple, Dict, Iterable

import numpy as np
from sklearn.feature_extraction.text import TfidfVectorizer

from kazu.utils.caching import kazu_disk_cache
from kazu.database.in_memory_db import SynonymDatabase
from kazu.utils.utils import create_char_ngrams, create_word_ngrams, Singleton

logger = logging.getLogger(__name__)


def create_word_and_char_ngrams(
    s: str,
    words: Iterable[int] = (
        1,
        2,
    ),
    chars: Iterable[int] = (
        2,
        3,
    ),
) -> List[str]:
    """
    function to create char and word ngrams

    :param s: string to process
    :param words: create n words
    :param chars: create n chars
    :return: list of strings comprised of words and chars
    """
    strings = []
    for ngram_word in words:
        strings.extend(create_word_ngrams(s, ngram_word))
    for ngram_char in chars:
        strings.extend(create_char_ngrams(s, ngram_char))
    return strings


class TfIdfScorer(metaclass=Singleton):
    """
    This class manages a set of TFIDF models (via
    :class:`sklearn.feature_extraction.text.TfidfVectorizer`\\ ).

    It's a singleton, so that the models can be accessed in multiple locations without the need to
    load them into memory multiple times.
    """

    def __init__(self):
        self.synonym_db = SynonymDatabase()
        self.parser_to_vectorizer: Dict[str, TfidfVectorizer] = self.build_vectorizers()

    @kazu_disk_cache.memoize(ignore={0})
    def build_vectorizers(self) -> Dict[str, TfidfVectorizer]:
        result: Dict[str, TfidfVectorizer] = {}
        for parser_name in self.synonym_db.loaded_parsers:
            synonyms = self.synonym_db.get_all(parser_name).keys()
            vectoriser = TfidfVectorizer(lowercase=False, analyzer=create_word_and_char_ngrams)
            vectoriser.fit(synonyms)
            result[parser_name] = vectoriser
        return result

    def __call__(
        self, strings: List[str], matrix: np.ndarray, parser: str
    ) -> Iterable[Tuple[str, float]]:
        """
        Transform a list of strings with a parser-specific vectorizer and score against a matrix.

        :param strings:
        :param matrix:
        :param parser:
        :return: matching strings and their score sorted by best score
        """
        if len(strings) == 1:
            yield strings[0], 100.0
        else:
            mat = self.parser_to_vectorizer[parser].transform(strings)
            score_matrix = np.squeeze(-np.asarray(mat.dot(matrix.T).todense()))
            neighbours = score_matrix.argsort()
            for neighbour in neighbours:
                yield strings[neighbour], -score_matrix[neighbour]



import logging
import pickle
import uuid
from collections import defaultdict
from dataclasses import dataclass, asdict
from functools import partial
from pathlib import Path
from typing import Any, List, Dict, Union, Iterable, Tuple, Optional, Set, cast

import spacy
import srsly
from kazu.database.in_memory_db import SynonymDatabase
from spacy import Language
from spacy.matcher import PhraseMatcher, Matcher
from spacy.tokens import Span, SpanGroup, Doc, Token
from spacy.util import SimpleFrozenList

from kazu.data.data import CuratedTermBehaviour, CuratedTerm
from kazu.ontology_preprocessing.base import (
    OntologyParser,
)
from kazu.utils.grouping import sort_then_group
from kazu.utils.utils import PathLike

GENE = "gene"
DRUG = "drug"
ANATOMY = "anatomy"
DISEASE = "disease"
CELL_LINE = "cell_line"
CELL_TYPE = "cell_type"
ENTITY = "entity"

SPAN_KEY = "RAW_HITS"
MATCH_ID_SEP = ":::"

logger = logging.getLogger(__name__)


@dataclass
class OntologyMatcherConfig:
    span_key: str
    match_id_sep: str
    labels: List[str]
    parser_name_to_entity_type: Dict[str, str]


def _ontology_dict_setter(span: Span, value: Dict):
    # spacy's typing on the property says this has to be a Dict[str, Any]
    # but the __init__ typing says it can be a Dict[Any, Any]
    span.doc.user_data[span] = value  # type: ignore[index]


def _ontology_dict_getter(span: Span):
    # as above with types
    return span.doc.user_data.get(span, {})  # type: ignore[call-overload]


@Language.factory(
    "ontology_matcher",
    default_config={
        "span_key": SPAN_KEY,
        "match_id_sep": MATCH_ID_SEP,
    },
)
class OntologyMatcher:
    """String matching to synonyms.

    Core strict matching is done by Spacy's `PhraseMatcher <https://spacy.io/api/phrasematcher>`_.
    """

    def __init__(
        self,
        nlp: Language,
        name: str = "ontology_matcher",
        *,
        span_key: str = SPAN_KEY,
        match_id_sep: str = MATCH_ID_SEP,
        parser_name_to_entity_type: Dict[str, str],
    ):
        """

        :param span_key: the key for doc.spans to store the matches in
        """

        Span.set_extension(
            "ontology_dict_", force=True, getter=_ontology_dict_getter, setter=_ontology_dict_setter
        )
        self.nlp = nlp
        self.name = name
        self.cfg = OntologyMatcherConfig(
            span_key=span_key,
            match_id_sep=match_id_sep,
            labels=[],
            parser_name_to_entity_type=parser_name_to_entity_type,
        )
        # These will be defined when calling initialize
        self.strict_matcher: Optional[PhraseMatcher] = None
        self.lowercase_matcher: Optional[PhraseMatcher] = None
        self.tp_matchers, self.fp_matchers = None, None
        self.tp_coocc_dict, self.fp_coocc_dict = None, None

    @property
    def nr_strict_rules(self) -> int:
        if not self.strict_matcher:
            return 0
        else:
            return len(self.strict_matcher)

    @property
    def nr_lowercase_rules(self) -> int:
        if not self.lowercase_matcher:
            return 0
        else:
            return len(self.lowercase_matcher)

    @property
    def span_key(self) -> str:
        return self.cfg.span_key

    @property
    def match_id_sep(self) -> str:
        return self.cfg.match_id_sep

    @property
    def labels(self) -> List[str]:
        """The labels currently processed by this component."""
        return self.cfg.labels

    @property
    def parser_name_to_entity_type(self) -> Dict[str, str]:
        return self.cfg.parser_name_to_entity_type

    def set_labels(self, labels: Iterable[str]):
        self.cfg.labels = list(labels)
        self.set_context_matchers()

    def set_context_matchers(self):
        self.tp_matchers, self.fp_matchers = self._create_token_matchers()
        self.tp_coocc_dict, self.fp_coocc_dict = self._create_coocc_dicts()

    def filter_curations_for_ner(
        self, curations: Iterable[CuratedTerm], parser: OntologyParser
    ) -> Iterable[CuratedTerm]:
        """check curations are still represented in DB before they can be used for NER."""
        original_terms = defaultdict(set)
        inherited_terms = defaultdict(set)
        syn_db = SynonymDatabase()
        for curation in curations:
            if curation.behaviour is CuratedTermBehaviour.ADD_FOR_NER_AND_LINKING:
                original_terms[curation.curated_synonym].add(curation)
            elif curation.behaviour is CuratedTermBehaviour.INHERIT_FROM_SOURCE_TERM:
                inherited_terms[curation.source_term].add(curation)
            else:
                # not an ner behaviour, so move on to the next curation in the loop
                continue
        for curated_synonym, curation_set in original_terms.items():
            if len(curation_set) > 1:
                # note, this shouldn't happen, as it should be handled by the curation validation logic.
                # however, just in case this logic changes, we'll leave this here
                logger.warning(
                    "multiple curations detected for %s, %s", curated_synonym, curation_set
                )
                term_norm = next(iter(curation_set)).term_norm_for_linking(parser.entity_class)
                if term_norm not in syn_db.get_all(parser.name):
                    logger.warning(
                        "dictionary based NER needs an database entry for %s, %s, but none exists",
                        term_norm,
                        parser.name,
                    )
                    continue
            yield from curation_set
            yield from inherited_terms.pop(curated_synonym, set())
        if len(inherited_terms) > 0:
            logger.warning(
                "The following inherited curations were unmatched to a source term, and will be ignored for NER: %s",
                inherited_terms.values(),
            )

    def create_phrasematchers_using_curations(
        self, parsers: List[OntologyParser]
    ) -> Tuple[Optional[PhraseMatcher], Optional[PhraseMatcher]]:
        """Create Spacy `PhraseMatcher <https://spacy.io/api/phrasematcher>`_\\ s based on :class:`.CuratedTerm`\\ s.

        Curations are produced by :py:meth:`.OntologyParser.populate_databases`\\ method
        :param parsers:
        :return:
        """
        if self.strict_matcher is not None or self.lowercase_matcher is not None:
            logging.warning("Phrase matchers are being redefined - is this by intention?")

        self.set_labels(parser.entity_class for parser in parsers)
        strict_matcher = PhraseMatcher(self.nlp.vocab, attr="ORTH")
        lowercase_matcher = PhraseMatcher(self.nlp.vocab, attr="NORM")
        logger.info("ontology matcher build triggered. Forcing database repopulation.")
        for parser in parsers:
            parser_curations = parser.populate_databases(force=True, return_curations=True)
            if parser_curations is None:
                logger.warning(
                    "tried to create PhraseMatchers from Curations for parser %s, but none have been provided",
                    parser.name,
                )
                continue
            if len(parser_curations) == 0:
                logger.warning(
                    "tried to create PhraseMatchers from Curations for parser %s, but no Curations were produced",
                    parser.name,
                )
                continue

            curations_for_ner = set(self.filter_curations_for_ner(parser_curations, parser))

            # deduplicating match id's and patterns saves memory in the spacy pipeline
            match_ids_and_strings_cs = set()
            match_ids_and_strings_ci = set()
            for curation in curations_for_ner:
                # a curation can have different term_norms for different parsers,
                # since the string normalizer's output depends on the entity class.
                # Also, a curation may exist in multiple SynonymTerm.terms
                term_norm = curation.term_norm_for_linking(parser.entity_class)
                match_id = (
                    parser.name
                    + self.match_id_sep
                    + term_norm
                    + self.match_id_sep
                    + str(curation.mention_confidence.value)
                )
                if curation.case_sensitive:
                    match_ids_and_strings_cs.add(
                        (
                            match_id,
                            curation.curated_synonym,
                        )
                    )
                else:
                    match_ids_and_strings_ci.add(
                        (
                            match_id,
                            curation.curated_synonym.lower(),
                        )
                    )

            for match_id, match_str in match_ids_and_strings_cs:
                strict_matcher.add(match_id, [self.nlp.tokenizer(match_str)])
            for match_id, match_str in match_ids_and_strings_ci:
                lowercase_matcher.add(match_id, [self.nlp.tokenizer(match_str)])

        # only set the phrasematcher if we have any rules for them
        # this lets us skip running a phrasematcher if it has no rules when we come
        # to calling OntologyMatcher
        self.strict_matcher = strict_matcher if len(strict_matcher) != 0 else None
        self.lowercase_matcher = lowercase_matcher if len(lowercase_matcher) != 0 else None
        return self.strict_matcher, self.lowercase_matcher

    def __call__(self, doc: Doc) -> Doc:
        if self.nr_strict_rules == 0 and self.nr_lowercase_rules == 0:
            raise ValueError("there are no matcher rules configured!")

        # at least one phrasematcher will now be set.
        # normally, this will only be one: either a strict matcher if constructed by curated list,
        # or a lowercase matcher if constructed 'from scrach' using the parsers - currently just an
        # initial step in building a curation-based phrasematcher

        if self.strict_matcher is not None and self.lowercase_matcher is not None:
            matcher_res = set(self.strict_matcher(doc)).union(set(self.lowercase_matcher(doc)))
        elif self.strict_matcher is not None:
            matcher_res = set(self.strict_matcher(doc))
        elif self.lowercase_matcher is not None:
            matcher_res = set(self.lowercase_matcher(doc))
        else:
            # this isn't possible since if both of them were None,
            # the if clause at the start of this function would raise an error
            raise AssertionError()

        matches = cast(Set[Tuple[int, int, int]], matcher_res)
        spans = []
        for (start, end), matches_grp in sort_then_group(
            matches,
            key_func=lambda x: (
                x[1],
                x[2],
            ),
        ):
            data = defaultdict(set)
            for mat in matches_grp:
                parser_name, term_norm, confidence = self.nlp.vocab.strings.as_string(mat[0]).split(
                    self.match_id_sep, maxsplit=2
                )
                ent_class = self.parser_name_to_entity_type[parser_name]
                data[ent_class].add((parser_name, term_norm, confidence))
            for ent_class, ent_class_data in data.items():
                # we use a uuid here so that every span hash is unique
                new_span = Span(doc, start, end, label=uuid.uuid4().hex)
                new_span._.set("ontology_dict_", {ent_class: ent_class_data})
                spans.append(new_span)
        final_spans = self.filter_by_contexts(doc, spans)
        span_group = SpanGroup(doc, name=self.span_key, spans=final_spans)
        doc.spans[self.span_key] = span_group
        return doc

    def filter_by_contexts(self, doc, spans):
        """These filters work best when there is sentence segmentation available."""
        doc_has_sents = doc.has_annotation("SENT_START")
        # set custom attributes for the token matchers
        for label in self.labels:
            Token.set_extension(label, default=False, force=True)
        # compile the filtered spans by going through each label and each span
        filtered_spans = []
        for label in self.labels:
            # Set all token attributes for other labels (to use in the matcher rules)
            for s in spans:
                for ent_class in s._.ontology_dict_:
                    if ent_class != label and ent_class in self.labels:
                        self._set_token_attr(ent_class=ent_class, span=s, value=True)
            # Set the token attribute for each span of this label one-by-one, then match
            for span in spans:
                for ent_class in span._.ontology_dict_:
                    if ent_class == label:
                        self._set_token_attr(ent_class=ent_class, span=span, value=True)
                        context = doc
                        if doc_has_sents:
                            context = span.sent
                        if (
                            self.span_in_TP_coocc(context, span=span, ent_class=ent_class)
                            and not self.span_in_FP_coocc(context, span=span, ent_class=ent_class)
                            and self.span_in_TP_context(context, ent_class)
                            and not self.span_in_FP_context(context, ent_class)
                        ):
                            filtered_spans.append(span)
                        # reset for the next span within the same label
                        self._set_token_attr(ent_class, span, False)
            # reset for the next label
            for s in spans:
                for ent_class in s._.ontology_dict_:
                    if ent_class in self.labels:
                        self._set_token_attr(ent_class, s, False)
        return filtered_spans

    def _set_token_attr(self, ent_class: str, span: Span, value: bool):
        for token in span:
            token._.set(ent_class, value)

    def span_in_TP_context(self, doc, ent_class: str):
        """When an entity type has a TP matcher defined, it should match for this
        span to be regarded as a true hit."""
        assert self.tp_matchers is not None
        tp_matcher = self.tp_matchers.get(ent_class, None)
        if tp_matcher:
            rule_spans = tp_matcher(doc, as_spans=True)
            if len(rule_spans) > 0:
                return True
            return False
        return True

    def span_in_FP_context(self, doc, ent_class: str):
        """When an entity type has a FP matcher defined, spans that match
        are regarded as FPs."""
        assert self.fp_matchers is not None
        fp_matcher = self.fp_matchers.get(ent_class, None)
        if fp_matcher:
            rule_spans = fp_matcher(doc, as_spans=True)
            if len(rule_spans) > 0:
                return True
            return False
        return False

    def span_in_TP_coocc(self, doc, span: Span, ent_class: str):
        """When an entity type has a TP co-occ dict defined, a hit defined in the dict
        is only regarded as a true hit when it matches at least one of its co-occ terms."""
        assert self.tp_coocc_dict is not None
        tp_dict = self.tp_coocc_dict.get(ent_class, None)
        if tp_dict and tp_dict.get(span.text):
            for w in tp_dict[span.text]:
                if w in doc.text:
                    return True
            return False
        return True

    def span_in_FP_coocc(self, doc, span: Span, ent_class: str):
        """When an entity type has a FP co-occ dic defined, a hit defined in the dict
        is regarded as a false positive when it matches at least one of its co-occ terms."""
        assert self.fp_coocc_dict is not None
        fp_dict = self.fp_coocc_dict.get(ent_class, None)
        if fp_dict and fp_dict.get(span.text):
            for w in fp_dict[span.text]:
                if w in doc.text:
                    return True
            return False
        return False

    def _create_token_matchers(self):
        tp_matchers = {}
        fp_matchers = {}
        if CELL_LINE in self.labels:
            tp_matchers[CELL_LINE] = self._create_cell_tp_tokenmatcher(CELL_LINE)
        if CELL_TYPE in self.labels:
            tp_matchers[CELL_TYPE] = self._create_cell_tp_tokenmatcher(CELL_TYPE)

        if ANATOMY in self.labels:
            fp_matchers[ANATOMY] = self._create_anatomy_fp_tokenmatcher()
        return tp_matchers, fp_matchers

    def _create_cell_tp_tokenmatcher(self, ent_class: str):
        """Define patterns where a Cell line or type appears and it's likely a true positive"""
        matcher = Matcher(self.nlp.vocab)
        pattern_1: List[Dict[str, Any]] = [
            {"_": {ent_class: True}},
            {"LOWER": {"IN": ["cell", "cells"]}},
        ]
        pattern_2: List[Dict[str, Any]] = [
            {"LOWER": "cell"},
            {"LOWER": "line"},
            {"_": {ent_class: True}},
        ]
        pattern_3: List[Dict[str, Any]] = [
            {"LOWER": "cell"},
            {"LOWER": "type"},
            {"_": {ent_class: True}},
        ]
        matcher.add("Cell_context", [pattern_1, pattern_2, pattern_3])
        return matcher

    def _create_anatomy_fp_tokenmatcher(self):
        """Define patterns where an atanomy entity appears and it's likely a false positive"""
        matcher = Matcher(self.nlp.vocab)
        patterns: List[List[Dict[str, Any]]] = []
        if DRUG in self.labels:
            p: List[Dict[str, Any]] = [{"_": {DRUG: True}}, {"_": {ANATOMY: True}, "LOWER": "arm"}]
            patterns.append(p)
        p2: List[Dict[str, Any]] = [
            {"LOWER": "single"},
            {"LOWER": "-"},
            {"_": {ANATOMY: True}, "LOWER": "arm"},
        ]
        patterns.append(p2)
        p3: List[Dict[str, Any]] = [
            {"LOWER": "quality"},
            {"LOWER": "-", "OP": "?"},
            {"LOWER": "of"},
            {"LOWER": "-", "OP": "?"},
            {"_": {ANATOMY: True}, "LOWER": "life"},
        ]
        patterns.append(p3)
        matcher.add("Anatomy_context", patterns)
        return matcher

    def _create_coocc_dicts(self):
        tp_coocc_dict: Dict[str, Dict[str, List[str]]] = {}
        fp_coocc_dict: Dict[str, Dict[str, List[str]]] = {}
        if GENE in self.labels:
            fp_coocc_dict[GENE] = self._create_gene_fp_dict()
        if DISEASE in self.labels:
            fp_coocc_dict[DISEASE] = self._create_disease_fp_dict()
        return tp_coocc_dict, fp_coocc_dict

    _ivf_fertility_treatment_cooccurrence = ["ICSI", "cycle", "treatment"]

    def _create_gene_fp_dict(self):
        """Define cooccurrence links that determine likely FP gene hits"""
        gene_dict = {}
        gene_dict["IVF"] = self._ivf_fertility_treatment_cooccurrence
        return gene_dict

    def _create_disease_fp_dict(self):
        """Define cooccurrence links that determine likely FP disease hits"""
        disease_dict = {}
        disease_dict["MFS"] = ["endpoint"]
        disease_dict["IVF"] = self._ivf_fertility_treatment_cooccurrence
        return disease_dict

    def to_disk(
        self, path: Union[str, Path], *, exclude: Iterable[str] = SimpleFrozenList()
    ) -> None:
        """Serialize the pipe to disk.

        :param path: Path to serialize the pipeline to.
        :param exclude: String names of serialization fields to exclude.
        """

        def pickle_matcher(p, strict):
            with p.open("wb") as outfile:
                if strict:
                    pickle.dump(self.strict_matcher, outfile)
                else:
                    pickle.dump(self.lowercase_matcher, outfile)

        serialize = {}
        serialize["cfg"] = lambda p: srsly.write_json(p, asdict(self.cfg))
        serialize["strict_matcher"] = partial(pickle_matcher, strict=True)
        serialize["lowercase_matcher"] = partial(pickle_matcher, strict=False)
        spacy.util.to_disk(path, serialize, exclude)

    def from_disk(
        self, path: Union[str, Path], *, exclude: Iterable[str] = SimpleFrozenList()
    ) -> "OntologyMatcher":
        """Load the pipe from disk. Modifies the object in place and returns it."""

        def unpickle_matcher(p, strict):
            with p.open("rb") as infile:
                matcher = pickle.load(infile)
                if strict:
                    self.strict_matcher = matcher
                else:
                    self.lowercase_matcher = matcher

        deserialize = {}

        def deserialize_cfg(path: PathLike) -> None:
            loaded_conf = srsly.read_json(path)
            self.cfg = OntologyMatcherConfig(**loaded_conf)

        deserialize["cfg"] = deserialize_cfg
        deserialize["strict_matcher"] = partial(unpickle_matcher, strict=True)
        deserialize["lowercase_matcher"] = partial(unpickle_matcher, strict=False)
        # spacy's typing says the functions have to only take strings, but they would normally take
        # a path as well without that being a problem.
        spacy.util.from_disk(path, deserialize, exclude)  # type: ignore[arg-type]
        self.set_context_matchers()
        return self



import functools
import logging
from abc import ABC, abstractmethod
from collections import defaultdict
from os import getenv
from typing import Tuple, Optional, Set, Dict, Iterable, FrozenSet

import numpy as np

from kazu.data.data import (
    Document,
    EquivalentIdSet,
    EquivalentIdAggregationStrategy,
    DisambiguationConfidence,
)
from kazu.database.in_memory_db import (
    MetadataDatabase,
    SynonymDatabase,
    NormalisedSynonymStr,
)
from kazu.steps.linking.post_processing.disambiguation.context_scoring import TfIdfScorer

logger = logging.getLogger(__name__)


class DisambiguationStrategy(ABC):
    """
    The job of a DisambiguationStrategy is to filter a Set of :class:`.EquivalentIdSet` into a
    (hopefully) smaller set.

    A :meth:`prepare` method is available, which can be cached in the event of any duplicated
    preprocessing work that may be required (see :class:`.StrategyRunner` for the complexities of
    how MappingStrategy and DisambiguationStrategy are coordinated).
    """

    def __init__(self, confidence: DisambiguationConfidence):
        """

        :param confidence: the level of confidence that should be assigned to this strategy. This is simply a label
            for human users, and has no bearing on the actual algorithm.
        """
        self.confidence = confidence

    @abstractmethod
    def prepare(self, document: Document):
        """
        perform any preprocessing required

        :param document:
        :return:
        """
        pass

    @abstractmethod
    def disambiguate(
        self, id_sets: Set[EquivalentIdSet], document: Document, parser_name: str
    ) -> Set[EquivalentIdSet]:
        """
        subset a set of :class:`.EquivalentIdSet`\\ .

        :param id_sets:
        :param document:
        :param parser_name:
        :return:
        """
        pass

    def __call__(
        self, id_sets: Set[EquivalentIdSet], document: Document, parser_name: str
    ) -> Set[EquivalentIdSet]:
        self.prepare(document)
        return self.disambiguate(id_sets, document, parser_name)


class DefinedElsewhereInDocumentDisambiguationStrategy(DisambiguationStrategy):
    """
    1. look for entities on the document that have mappings
    2. see if any of these mappings correspond to any ids in the :class:`.EquivalentIdSet` on each
       hit
    """

    def __init__(self, confidence: DisambiguationConfidence):
        super().__init__(confidence)
        self.mapped_ids: Set[Tuple[str, str, str]] = set()

    def prepare(self, document: Document):
        """
        note, this method can't be cached, as the state of the document may change between executions

        :param document:
        :return:
        """
        self.mapped_ids = set()
        entities = document.get_entities()
        self.mapped_ids.update(
            (
                mapping.parser_name,
                mapping.source,
                mapping.idx,
            )
            for ent in entities
            for mapping in ent.mappings
        )

    def disambiguate(
        self, id_sets: Set[EquivalentIdSet], document: Document, parser_name: str
    ) -> Set[EquivalentIdSet]:
        found_id_sets = set()
        for id_set in id_sets:
            for idx, source in id_set.ids_and_source:
                if (
                    parser_name,
                    source,
                    idx,
                ) in self.mapped_ids:
                    found_id_sets.add(id_set)
                    break
        return found_id_sets


class TfIdfDisambiguationStrategy(DisambiguationStrategy):
    """
    1. retrieve all synonyms associated with a :class:`.EquivalentIdSet`, filter out ambiguous ones
       and build a query matrix with the unambiguous ones.
    2. retrieve a list of all detected entity strings from the document, regardless of source and
       build a document representation matrix of these.
    3. perform TFIDF on the query vs document, and sort according to most likely synonym hit from 1.
    4. if the score is above the minimum threshold, create a mapping.

    """

    CONTEXT_SCORE = "context_score"

    def __init__(
        self,
        confidence: DisambiguationConfidence,
        scorer: TfIdfScorer,
        context_threshold: float = 0.7,
        relevant_aggregation_strategies: Optional[Iterable[EquivalentIdAggregationStrategy]] = None,
    ):
        """

        :param scorer: handles scoring of contexts
        :param context_threshold: only consider terms above this search threshold
        :param relevant_aggregation_strategies: Only consider these strategies when selecting synonyms from the
            synonym database, when building a representation. If none, all strategies will be considered
        """
        super().__init__(confidence)
        self.context_threshold = context_threshold
        if relevant_aggregation_strategies is None:
            self.relevant_aggregation_strategies = {EquivalentIdAggregationStrategy.UNAMBIGUOUS}
        else:
            self.relevant_aggregation_strategies = set(relevant_aggregation_strategies)
        self.synonym_db = SynonymDatabase()
        self.scorer = scorer
        self.parser_name_to_doc_representation: Dict[str, np.ndarray] = {}

    @functools.lru_cache(maxsize=int(getenv("KAZU_TFIDF_DISAMBIGUATION_DOCUMENT_CACHE_SIZE", 1)))
    def prepare(self, document: Document):
        """
        build document representations by parser names here, and store in a dict. This method is cached so
        we don't need to call it multiple times per document

        :param document:
        :return:
        """
        parser_names = frozenset(
            term.parser_name
            for ent in document.get_entities()
            for term in ent.syn_term_to_synonym_terms
        )
        self.parser_name_to_doc_representation = self.cacheable_build_document_representation(
            scorer=self.scorer, doc=document, parsers=parser_names
        )

    @staticmethod
    @functools.lru_cache(maxsize=int(getenv("KAZU_TFIDF_DISAMBIGUATION_CACHE_SIZE", 20)))
    def cacheable_build_document_representation(
        scorer: TfIdfScorer, doc: Document, parsers: FrozenSet[str]
    ) -> Dict[str, np.ndarray]:
        """
        static cached method, so we don't need to recalculate document representation between different instances
        of this class.

        :param scorer:
        :param doc:
        :param parsers: technically this only has to be a hashable iterable of string - but it should also be
            unique otherwise duplicate work will be done and thrown away, so pragmatically a frozenset makes sense.
        :return:
        """
        strings = " ".join(x.match_norm for x in doc.get_entities())
        res = {}
        for parser in parsers:
            vectorizer = scorer.parser_to_vectorizer[parser]
            res[parser] = vectorizer.transform([strings])
        return res

    def build_id_set_representation(
        self,
        parser_name: str,
        id_sets: Set[EquivalentIdSet],
    ) -> Dict[NormalisedSynonymStr, Set[EquivalentIdSet]]:
        result = defaultdict(set)
        for id_set in id_sets:
            for idx in id_set.ids:

                syns_this_id = self.synonym_db.get_syns_for_id(
                    parser_name,
                    idx,
                    self.relevant_aggregation_strategies,
                )
                for syn in syns_this_id:
                    result[syn].add(id_set)
        return result

    def disambiguate(
        self, id_sets: Set[EquivalentIdSet], document: Document, parser_name: str
    ) -> Set[EquivalentIdSet]:
        if parser_name not in self.scorer.parser_to_vectorizer:
            return set()

        document_query_matrix = self.parser_name_to_doc_representation[parser_name]
        id_set_representation = self.build_id_set_representation(parser_name, id_sets)
        if len(id_set_representation) == 0:
            return set()
        else:
            indexed_non_ambiguous_syns = list(id_set_representation.keys())
            for best_syn, score in self.scorer(
                strings=indexed_non_ambiguous_syns,
                matrix=document_query_matrix,
                parser=parser_name,
            ):
                if score >= self.context_threshold and len(id_set_representation[best_syn]) == 1:
                    return id_set_representation[best_syn]
            return set()


class AnnotationLevelDisambiguationStrategy(DisambiguationStrategy):
    """
    certain entities are often mentioned by some colloquial name, even if it's technically incorrect. In these cases,
    we may have an annotation_score field in the metadata_db, as a proxy of how widely studied the entity is. We
    use this annotation score as a proxy for 'given a random mention of the entity, how likely is it that the
    author is referring to instance x vs instance y'. Naturally, this is a pretty unsophisticated disambiguation
    strategy, so should generally only be used as a last resort!
    """

    def prepare(self, document: Document):
        pass

    def disambiguate(
        self, id_sets: Set[EquivalentIdSet], document: Document, parser_name: str
    ) -> Set[EquivalentIdSet]:
        best_score = 0
        best_equiv_id_sets = set()

        for id_set in id_sets:
            for idx in id_set.ids:
                score = int(
                    MetadataDatabase().get_by_idx(parser_name, idx).get("annotation_score", 0)
                )
                if score > best_score:
                    best_score = score
                    best_equiv_id_sets = {id_set}
                elif score == best_score:
                    best_equiv_id_sets.add(id_set)

        return best_equiv_id_sets



import os
from pathlib import Path
from typing import Any, Dict, Optional, Union

import torch
from omegaconf import OmegaConf
from pytorch_lightning.plugins import CheckpointIO
from transformers import AutoTokenizer


class StudentModelCheckpointIO(CheckpointIO):
    """
    A plugin for saving student model (without saving teacher model)
    """

    def __init__(self, model_name_or_path: str):
        super().__init__()
        self.model_name_or_path = model_name_or_path

    def save_checkpoint(
        self,
        checkpoint: Dict[str, Any],
        path: Union[str, Path],
        storage_options: Optional[Any] = None,
    ) -> None:
        """
        Save distilled (student) model. Loading currently not implemented

        :param checkpoint: contents to save. Including `state_dict`, `optimizer_states` and `callbacks`.
        :param path:
        :param storage_options:
        """

        dirPath = os.path.dirname(path)

        output_config_file = os.path.join(dirPath, "hyper_parameters.json")
        OmegaConf.save(
            OmegaConf.create(checkpoint["hyper_parameters"]), output_config_file, resolve=True
        )
        studentModel_state_dict = {
            key[len("student_model.") :]: value
            for key, value in checkpoint["state_dict"].items()
            if key.startswith("student_model.")
        }
        teacherModel_state_dict = {
            key: value
            for key, value in checkpoint["state_dict"].items()
            if key.startswith("teacher_model.")
        }
        assert len(checkpoint["state_dict"]) == len(studentModel_state_dict) + len(
            teacherModel_state_dict
        ), "Missing structures while saving trained model."
        torch.save(studentModel_state_dict, path)

        AutoTokenizer.from_pretrained(self.model_name_or_path).save_vocabulary(
            os.path.dirname(path)
        )

    def load_checkpoint(
        self, path: Union[str, Path], storage_options: Optional[Any] = None
    ) -> Dict[str, Any]:
        """Not currently implemented.

        See :external+pytorch_lightning:meth:`CheckpointIO.load_checkpoint <lightning.pytorch.plugins.io.CheckpointIO.load_checkpoint>`
        for details of the abstract method."""
        raise NotImplementedError

    def remove_checkpoint(
        self,
        path: Union[str, Path],
    ) -> None:

        os.remove(path)



"""Modified for distillation using Pytorch Lightning by KAZU team.

Based off of the `DataProcessor <https://github.com/dmis-lab/biobert/blob/master/run_ner.py#L127>`_
and `NerProcessor <https://github.com/dmis-lab/biobert/blob/master/run_ner.py#L175>`_ classes in BioBERT,
also written in reference to Huawei Noah's Ark Lab `TinyBERT <https://github.com/huawei-noah/Pretrained-Language-Model/blob/master/TinyBERT>`_.

Licensed under Apache 2.0

| Copyright 2018 The Google AI Language Team Authors, The HuggingFace Inc. team and Huawei Noah's Ark Lab.
| Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.

.. raw:: html

    <details>
    <summary>Full License Notice</summary>

| Copyright 2018 The Google AI Language Team Authors, The HuggingFace Inc. team and Huawei Noah's Ark Lab.
| Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

.. raw:: html

    </details>
"""

import logging
import os
from typing import List, Tuple, Iterable

from transformers import InputExample, DataProcessor

from kazu.utils.utils import PathLike

logger = logging.getLogger(__name__)


class SeqTagProcessor(DataProcessor):
    """Base class for data converters for sequence tagging data sets."""

    def get_train_examples(self, data_dir) -> List[InputExample]:
        """Gets a collection of :class:`transformers.InputExample` for the train set."""
        raise NotImplementedError

    def get_dev_examples(self, data_dir) -> List[InputExample]:
        """Gets a collection of :class:`transformers.InputExample` for the dev set."""
        raise NotImplementedError

    def get_aug_examples(self, data_dir) -> List[InputExample]:
        """Gets a collection of :class:`transformers.InputExample` for the aug set."""
        raise NotImplementedError


class NerProcessor(SeqTagProcessor):
    def get_train_examples(self, data_dir: str) -> List[InputExample]:
        return self._create_examples(
            self._read_data(os.path.join(data_dir, "train_dev.tsv")), "train"
        )

    def get_dev_examples(self, data_dir: str) -> List[InputExample]:
        return self._create_examples(self._read_data(os.path.join(data_dir, "devel.tsv")), "dev")

    def get_test_examples(self, data_dir: str) -> List[InputExample]:
        """Gets a collection of :class:`transformers.InputExample` for the test set."""
        return self._create_examples(self._read_data(os.path.join(data_dir, "test.tsv")), "test")

    def get_aug_examples(self, data_dir: str) -> List[InputExample]:
        return self._create_examples(self._read_tsv(os.path.join(data_dir, "train_aug.tsv")), "aug")

    def _create_examples(
        self, lines: Iterable[Tuple[str, str]], set_type: str
    ) -> List[InputExample]:
        examples = []
        for (i, line) in enumerate(lines):
            guid = "%s-%s" % (set_type, i)
            # TODO assert if tokenization from BERT and tokenization from pytorch mismatch
            text = line[1]
            label = line[0]
            examples.append(InputExample(guid=guid, text_a=text, label=label))
        return examples

    @classmethod
    def _read_data(cls, input_file: PathLike) -> List[Tuple[str, str]]:
        """Reads a BIO data."""
        with open(input_file) as inpFilept:
            lines = []
            words: List[str] = []
            labels: List[str] = []
            continualLineErrorCnt = 0
            for lineIdx, line in enumerate(inpFilept):
                contents = line.splitlines()[0]
                lineList = contents.split()
                if len(lineList) == 0:  # For blank line
                    assert len(words) == len(
                        labels
                    ), "lineIdx: %s,  len(words)(%s) != len(labels)(%s) \n %s\n%s" % (
                        lineIdx,
                        len(words),
                        len(labels),
                        " ".join(words),
                        " ".join(labels),
                    )
                    if len(words) != 0:
                        wordSent = " ".join(words)
                        labelSent = " ".join(labels)
                        lines.append((labelSent, wordSent))
                        words = []
                        labels = []
                    else:
                        continualLineErrorCnt += 1
                else:
                    words.append(lineList[0])
                    labels.append(lineList[-1])
            if len(words) != 0:
                wordSent = " ".join(words)
                labelSent = " ".join(labels)
                lines.append((labelSent, wordSent))

        logging.info("continualLineErrorCnt : %s" % (continualLineErrorCnt))
        return lines



from typing import List

try:
    from seqeval.metrics import f1_score
except ImportError as e:
    raise ImportError(
        "Running the model distillation code requires seqeval to be installed.\n"
        "We recommend running 'pip install kazu[model_training]' to get all model training"
        " dependencies."
    ) from e


IGNORE_IDX = -100


def accuracy(preds, labels):
    return (preds == labels).mean()


def numeric_label_f1_score(preds: List[List[int]], golds: List[List[int]], label_list) -> float:
    """
    Function to calculate F1 score using seqeval and numerical format labels.

    :param preds: 2d array of predicted label ids
    :param golds: 2d array of gold standard ids
    :param label_list: list of strings, for mappingids to labels
    :return:
    """

    pred_clean_labels_list = []
    gold_clean_labels_list = []

    assert len(preds) == len(golds)
    for preds_id_sequence, golds_id_sequence in zip(preds, golds):
        assert len(preds_id_sequence) == len(golds_id_sequence)
        p_labels = []
        g_labels = []
        for pred_label, gold_label in zip(preds_id_sequence, golds_id_sequence):
            if gold_label != IGNORE_IDX:
                p_labels.append(label_list[pred_label])
                g_labels.append(label_list[gold_label])
        pred_clean_labels_list.append(p_labels)
        gold_clean_labels_list.append(g_labels)

    return f1_score(gold_clean_labels_list, pred_clean_labels_list)



import re
from collections import Counter
from typing import Protocol, List

from cachetools import LFUCache
from pytorch_lightning import Trainer
from rapidfuzz import fuzz
from torch import Tensor, cosine_similarity

from kazu.data.data import NumericMetric
from kazu.linking.sapbert.train import PLSapbertModel
from kazu.utils.utils import Singleton


class StringSimilarityScorer(Protocol):
    """
    calculates a NumericMetric based on a string match or a normalised string match and a normalised term
    """

    def __call__(self, reference_term: str, query_term: str) -> NumericMetric:
        raise NotImplementedError


class BooleanStringSimilarityScorer(StringSimilarityScorer, Protocol):
    def __call__(self, reference_term: str, query_term: str) -> bool:
        raise NotImplementedError


class NumberMatchStringSimilarityScorer(BooleanStringSimilarityScorer):
    """
    checks all numbers in reference_term are represented in term_norm
    """

    number_finder = re.compile("[0-9]+")

    @classmethod
    def __call__(cls, reference_term: str, query_term: str) -> bool:
        reference_term_number_count = Counter(cls.number_finder.findall(reference_term))
        query_term_number_count = Counter(cls.number_finder.findall(query_term))
        return reference_term_number_count == query_term_number_count


class EntitySubtypeStringSimilarityScorer(BooleanStringSimilarityScorer):
    """
    checks all TYPE x mentions in match norm are represented in term norm
    """

    # need to handle I explicitly
    # other roman numerals get normalized to integers,
    # but not I as this would be problematic
    numeric_class_phrases = re.compile("|".join(["TYPE (?:I|[0-9]+)"]))

    @classmethod
    def __call__(cls, reference_term: str, query_term: str) -> bool:
        reference_term_numeric_phrase_count = Counter(
            cls.numeric_class_phrases.findall(reference_term)
        )
        query_term_numeric_phrase_count = Counter(cls.numeric_class_phrases.findall(query_term))

        # we don't want to just do reference_term_numeric_phrase_count == query_term_numeric_phrase_count
        # because e.g. if reference term is 'diabetes' that is an NER match we've picked up in some text,
        # we want to keep hold of all of 'diabetes type I', 'diabetes type II', 'diabetes', in case surrounding context
        # enables us to disambiguate which type of diabetes it is
        return all(
            numeric_class_phase in query_term_numeric_phrase_count
            and query_term_numeric_phrase_count[numeric_class_phase] >= count
            for numeric_class_phase, count in reference_term_numeric_phrase_count.items()
        )


class EntityNounModifierStringSimilarityScorer(BooleanStringSimilarityScorer):
    """
    checks all modifier phrases in reference_term are represented in term_norm
    """

    def __init__(self, noun_modifier_phrases: List[str]):
        self.noun_modifier_phrases = noun_modifier_phrases

    def __call__(self, reference_term: str, query_term: str) -> bool:
        # the pattern should either be in both or neither
        return all(
            (pattern in reference_term) == (pattern in query_term)
            for pattern in self.noun_modifier_phrases
        )


class RapidFuzzStringSimilarityScorer(StringSimilarityScorer):
    """
    uses rapid fuzz to calculate string similarity. Note, if the token count >4 and reference_term has
    more than 10 chars, token_sort_ratio is used. Otherwise, WRatio is used
    """

    @staticmethod
    def __call__(reference_term: str, query_term: str) -> NumericMetric:
        if len(reference_term) > 10 and len(reference_term.split(" ")) > 4:
            return fuzz.token_sort_ratio(reference_term, query_term)
        else:
            return fuzz.WRatio(reference_term, query_term)


class SapbertStringSimilarityScorer(metaclass=Singleton):
    """
    note this is an implementation of the StringSimilarityScorer Protocol, but as a Singleton we can't inherit it
    """

    def __init__(self, sapbert: PLSapbertModel, trainer: Trainer, cache_size: int = 1000):
        """

        :param sapbert: model to use
        :param trainer: trainer to use
        :param cache_size: cache size, to prevent repeated calls to sapbert for the same string
        """
        self.trainer = trainer
        self.sapbert = sapbert
        self.embedding_cache: LFUCache[str, Tensor] = LFUCache(maxsize=cache_size)

    def __call__(self, reference_term: str, query_term: str) -> float:
        if reference_term == query_term:
            return 1.0

        ref_embedding = self.embedding_cache.get(reference_term)
        query_embedding = self.embedding_cache.get(query_term)
        if ref_embedding is None and query_embedding is None:
            embeddings = self.sapbert.get_embeddings_for_strings(
                [reference_term, query_term], batch_size=2, trainer=self.trainer
            )
            ref_embedding = embeddings[0]
            query_embedding = embeddings[1]
            self.embedding_cache[reference_term] = ref_embedding
            self.embedding_cache[query_term] = query_embedding
        elif ref_embedding is None:
            embeddings = self.sapbert.get_embeddings_for_strings(
                [reference_term], batch_size=1, trainer=self.trainer
            )
            ref_embedding = embeddings[0]
            self.embedding_cache[reference_term] = ref_embedding
        elif query_embedding is None:
            embeddings = self.sapbert.get_embeddings_for_strings(
                [query_term], batch_size=1, trainer=self.trainer
            )
            query_embedding = embeddings[0]
            self.embedding_cache[query_term] = query_embedding

        assert ref_embedding is not None
        assert query_embedding is not None
        return cosine_similarity(ref_embedding, query_embedding, dim=0).item()



import json
import logging
import os
import re
import sqlite3
from functools import cache
from pathlib import Path
from typing import (
    cast,
    List,
    Tuple,
    Dict,
    Any,
    Iterable,
    Set,
    Optional,
    Union,
    overload,
)
from urllib import parse

import pandas as pd
import rdflib

from kazu.data.data import (
    EquivalentIdSet,
    EquivalentIdAggregationStrategy,
    CuratedTerm,
    AssociatedIdSets,
    GlobalParserActions,
)
from kazu.language.string_similarity_scorers import StringSimilarityScorer
from kazu.ontology_preprocessing.base import (
    OntologyParser,
    SYN,
    MAPPING_TYPE,
    DEFAULT_LABEL,
    IDX,
    IdsAndSource,
)
from kazu.ontology_preprocessing.synonym_generation import CombinatorialSynonymGenerator
from kazu.utils.utils import PathLike

logger = logging.getLogger(__name__)


class JsonLinesOntologyParser(OntologyParser):
    """
    A parser for a jsonlines dataset. Assumes one kb entry per line (i.e. json object)
    implemetations should implement json_dict_to_parser_dict (see method notes for details
    """

    def read(self, path: str) -> Iterable[Dict[str, Any]]:
        for json_path in Path(path).glob("*.json"):
            with json_path.open(mode="r") as f:
                for line in f:
                    yield json.loads(line)

    def parse_to_dataframe(self):
        return pd.DataFrame.from_records(self.json_dict_to_parser_records(self.read(self.in_path)))

    def json_dict_to_parser_records(
        self, jsons_gen: Iterable[Dict[str, Any]]
    ) -> Iterable[Dict[str, Any]]:
        """
        for a given input json (represented as a python dict), yield dictionary record(s) compatible with the expected
        structure of the Ontology Parser superclass - i.e. should have keys for SYN, MAPPING_TYPE, DEFAULT_LABEL and
        IDX. All other keys are used as mapping metadata

        :param jsons_gen: iterator of python dict representing json objects
        :return:
        """
        raise NotImplementedError


class OpenTargetsDiseaseOntologyParser(JsonLinesOntologyParser):
    # Just use IDs that are in MONDO, since that's all people in general care about.
    # if we want to expand this out, other sources are:
    # "OGMS", "FBbt", "Orphanet", "EFO", "OTAR"
    # but we did have these in previously, and EFO introduced a lot of noise as it
    # has non-disease terms like 'dose' that occur frequently.
    # we could make the allowed sources a config option but we don't need to configure
    # currently, and easy to change later (and provide the current value as a default if
    # not present in config)
    allowed_sources = {"MONDO", "HP"}

    def find_kb(self, string: str) -> str:
        return string.split("_")[0]

    def json_dict_to_parser_records(
        self, jsons_gen: Iterable[Dict[str, Any]]
    ) -> Iterable[Dict[str, Any]]:
        # we ignore related syns for now until we decide how the system should handle them
        for json_dict in jsons_gen:
            idx = self.look_for_mondo(json_dict["id"], json_dict.get("dbXRefs", []))
            if any(allowed_source in idx for allowed_source in self.allowed_sources):
                synonyms = json_dict.get("synonyms", {})
                exact_syns = synonyms.get("hasExactSynonym", [])
                exact_syns.append(json_dict["name"])
                def_label = json_dict["name"]
                dbXRefs = json_dict.get("dbXRefs", [])
                for syn in exact_syns:
                    yield {
                        SYN: syn,
                        MAPPING_TYPE: "hasExactSynonym",
                        DEFAULT_LABEL: def_label,
                        IDX: idx,
                        "dbXRefs": dbXRefs,
                    }

    def look_for_mondo(self, ot_id: str, db_xrefs: List[str]):
        if "MONDO" in ot_id:
            return ot_id
        for x in db_xrefs:
            if "MONDO" in x:
                return x.replace(":", "_")
        return ot_id


class OpenTargetsTargetOntologyParser(JsonLinesOntologyParser):

    annotation_fields = {
        "subcellularLocations",
        "tractability",
        "constraint",
        "functionDescriptions",
        "go",
        "hallmarks",
        "chemicalProbes",
        "safetyLiabilities",
        "pathways",
        "targetClass",
    }

    def score_and_group_ids(
        self,
        ids_and_source: IdsAndSource,
        is_symbolic: bool,
        original_syn_set: Set[str],
    ) -> Tuple[AssociatedIdSets, EquivalentIdAggregationStrategy]:
        """
        since non symbolic gene symbols are also frequently ambiguous, we override this method accordingly to disable
        all synonym resolution, and rely on disambiguation to decide on 'true' mappings. Answers on a postcard if anyone
        has a better idea on how to do this!

        :param id_and_source:
        :param is_symbolic:
        :param original_syn_set:
        :return:
        """

        return (
            frozenset(
                EquivalentIdSet(ids_and_source=frozenset((single_id_and_source,)))
                for single_id_and_source in ids_and_source
            ),
            EquivalentIdAggregationStrategy.CUSTOM,
        )

    def find_kb(self, string: str) -> str:
        return "ENSEMBL"

    def json_dict_to_parser_records(
        self, jsons_gen: Iterable[Dict[str, Any]]
    ) -> Iterable[Dict[str, Any]]:
        for json_dict in jsons_gen:
            # due to a bug in OT data, TEC genes have "gene" as a synonym. Sunce they're uninteresting, we just filter
            # them
            biotype = json_dict.get("biotype")
            if biotype == "" or biotype == "tec" or json_dict["id"] == json_dict["approvedSymbol"]:
                continue

            annotation_score = sum(
                1
                for annotation_field in self.annotation_fields
                if len(json_dict.get(annotation_field, [])) > 0
            )

            shared_values = {
                IDX: json_dict["id"],
                DEFAULT_LABEL: json_dict["approvedSymbol"],
                "dbXRefs": json_dict.get("dbXRefs", []),
                "approvedName": json_dict["approvedName"],
                "annotation_score": annotation_score,
            }

            for key in ["synonyms", "obsoleteSymbols", "obsoleteNames", "proteinIds"]:
                synonyms_and_sources_lst = json_dict.get(key, [])
                for record in synonyms_and_sources_lst:
                    if "label" in record and "id" in record:
                        raise RuntimeError(f"record: {record} has both id and label specified")
                    elif "label" in record:
                        record[SYN] = record.pop("label")
                    elif "id" in record:
                        record[SYN] = record.pop("id")
                    record[MAPPING_TYPE] = record.pop("source")
                    record.update(shared_values)
                    yield record

            for key in ("approvedSymbol", "approvedName", "id"):
                if key == "id":
                    mapping_type = "opentargets_id"
                else:
                    mapping_type = key

                res = {SYN: json_dict[key], MAPPING_TYPE: mapping_type}
                res.update(shared_values)
                yield res


class OpenTargetsMoleculeOntologyParser(JsonLinesOntologyParser):
    def find_kb(self, string: str) -> str:
        return "CHEMBL"

    def json_dict_to_parser_records(
        self, jsons_gen: Iterable[Dict[str, Any]]
    ) -> Iterable[Dict[str, Any]]:
        for json_dict in jsons_gen:
            cross_references = json_dict.get("crossReferences", {})
            default_label = json_dict["name"]
            idx = json_dict["id"]

            synonyms = json_dict.get("synonyms", [])
            main_name = json_dict["name"]
            synonyms.append(main_name)

            for syn in synonyms:
                yield {
                    SYN: syn,
                    MAPPING_TYPE: "synonyms",
                    "crossReferences": cross_references,
                    DEFAULT_LABEL: default_label,
                    IDX: idx,
                }

            for trade_name in json_dict.get("tradeNames", []):
                yield {
                    SYN: trade_name,
                    MAPPING_TYPE: "tradeNames",
                    "crossReferences": cross_references,
                    DEFAULT_LABEL: default_label,
                    IDX: idx,
                }


RdfRef = Union[rdflib.paths.Path, rdflib.term.Node, str]
# Note - lists are actually normally provided here through hydra config
# but there's apparently no way of type hinting
# 'any iterable of length two where the items have these types'
PredicateAndValue = Tuple[RdfRef, rdflib.term.Node]


class RDFGraphParser(OntologyParser):
    """
    Parser for Owl files.
    """

    def __init__(
        self,
        in_path: str,
        entity_class: str,
        name: str,
        uri_regex: Union[str, re.Pattern],
        synonym_predicates: Iterable[RdfRef],
        string_scorer: Optional[StringSimilarityScorer] = None,
        synonym_merge_threshold: float = 0.7,
        data_origin: str = "unknown",
        synonym_generator: Optional[CombinatorialSynonymGenerator] = None,
        include_entity_patterns: Optional[Iterable[PredicateAndValue]] = None,
        exclude_entity_patterns: Optional[Iterable[PredicateAndValue]] = None,
        curations_path: Optional[str] = None,
        global_actions: Optional[GlobalParserActions] = None,
        label_predicate: RdfRef = rdflib.RDFS.label,
    ):
        super().__init__(
            in_path=in_path,
            entity_class=entity_class,
            name=name,
            string_scorer=string_scorer,
            synonym_merge_threshold=synonym_merge_threshold,
            data_origin=data_origin,
            synonym_generator=synonym_generator,
            curations_path=curations_path,
            global_actions=global_actions,
        )

        if isinstance(uri_regex, re.Pattern):
            self._uri_regex = uri_regex
        else:
            self._uri_regex = re.compile(uri_regex)

        self.synonym_predicates = tuple(
            self.convert_to_rdflib_ref(pred) for pred in synonym_predicates
        )
        self.label_predicate = self.convert_to_rdflib_ref(label_predicate)

        if include_entity_patterns is not None:
            self.include_entity_patterns = tuple(
                (self.convert_to_rdflib_ref(pred), self.convert_to_rdflib_ref(val))
                for pred, val in include_entity_patterns
            )
        else:
            self.include_entity_patterns = tuple()

        if exclude_entity_patterns is not None:
            self.exclude_entity_patterns = tuple(
                (self.convert_to_rdflib_ref(pred), self.convert_to_rdflib_ref(val))
                for pred, val in exclude_entity_patterns
            )
        else:
            self.exclude_entity_patterns = tuple()

    def find_kb(self, string: str) -> str:
        # By default, just return the name of the parser.
        # If more complex behaviour is necessary, write a custom subclass and override this method.
        return self.name

    @overload
    @staticmethod
    def convert_to_rdflib_ref(pred: rdflib.paths.Path) -> rdflib.paths.Path:
        pass

    @overload
    @staticmethod
    def convert_to_rdflib_ref(pred: rdflib.term.Node) -> rdflib.term.Node:
        pass

    @overload
    @staticmethod
    def convert_to_rdflib_ref(pred: str) -> rdflib.URIRef:
        pass

    @staticmethod
    def convert_to_rdflib_ref(pred):
        if isinstance(pred, (rdflib.term.Node, rdflib.paths.Path)):
            return pred
        else:
            return rdflib.URIRef(pred)

    def parse_to_dataframe(self) -> pd.DataFrame:
        g = rdflib.Graph()
        g.parse(self.in_path)
        default_labels = []
        iris = []
        syns = []
        mapping_type = []

        label_pred_str = str(self.label_predicate)

        for sub, obj in g.subject_objects(self.label_predicate):
            if not self.is_valid_iri(str(sub)):
                continue

            # type ignore is necessary because rdflib's typing thinks that for Graph.__contains__ can't use an rdflib.paths.Path
            # as a predicate, but you can, because __contains__ calls Graph.triples(), which is type hinted to allow Paths (and
            # reading the implementation it clearly handles Paths).
            if any((sub, pred, value) not in g for pred, value in self.include_entity_patterns):  # type: ignore[operator]
                continue

            # as above
            if any((sub, pred, value) in g for pred, value in self.exclude_entity_patterns):  # type: ignore[operator]
                continue

            default_labels.append(str(obj))
            iris.append(str(sub))
            syns.append(str(obj))
            mapping_type.append(label_pred_str)
            for syn_predicate in self.synonym_predicates:
                for other_syn_obj in g.objects(subject=sub, predicate=syn_predicate):
                    default_labels.append(str(obj))
                    iris.append(str(sub))
                    syns.append(str(other_syn_obj))
                    mapping_type.append(str(syn_predicate))

        df = pd.DataFrame.from_dict(
            {DEFAULT_LABEL: default_labels, IDX: iris, SYN: syns, MAPPING_TYPE: mapping_type}
        )
        return df

    def is_valid_iri(self, text: str) -> bool:
        """
        Check if input string is a valid IRI for the ontology being parsed.

        Uses `self._uri_regex` to define valid IRIs
        """
        match = self._uri_regex.match(text)
        return bool(match)


SKOS_XL_PREF_LABEL_PATH: rdflib.paths.Path = rdflib.URIRef(
    "http://www.w3.org/2008/05/skos-xl#prefLabel"
) / rdflib.URIRef("http://www.w3.org/2008/05/skos-xl#literalForm")
SKOS_XL_ALT_LABEL_PATH: rdflib.paths.Path = rdflib.URIRef(
    "http://www.w3.org/2008/05/skos-xl#altLabel"
) / rdflib.URIRef("http://www.w3.org/2008/05/skos-xl#literalForm")


class SKOSXLGraphParser(RDFGraphParser):
    """Parse SKOS-XL RDF Files.

    Note that this just sets a default label predicate and synonym predicate to SKOS-XL
    appropriate paths, and then passes to the parent RDFGraphParser class. This class is just a convenience
    to make specifying a SKOS-XL parser easier, this functionality is still available via RDFGraphParser
    directly.
    """

    def __init__(
        self,
        in_path: str,
        entity_class: str,
        name: str,
        uri_regex: Union[str, re.Pattern],
        synonym_predicates: Iterable[RdfRef] = (SKOS_XL_ALT_LABEL_PATH,),
        string_scorer: Optional[StringSimilarityScorer] = None,
        synonym_merge_threshold: float = 0.7,
        data_origin: str = "unknown",
        synonym_generator: Optional[CombinatorialSynonymGenerator] = None,
        include_entity_patterns: Optional[Iterable[PredicateAndValue]] = None,
        exclude_entity_patterns: Optional[Iterable[PredicateAndValue]] = None,
        curations_path: Optional[str] = None,
        global_actions: Optional[GlobalParserActions] = None,
        label_predicate: RdfRef = SKOS_XL_PREF_LABEL_PATH,
    ):
        super().__init__(
            in_path=in_path,
            entity_class=entity_class,
            name=name,
            uri_regex=uri_regex,
            synonym_predicates=synonym_predicates,
            string_scorer=string_scorer,
            synonym_merge_threshold=synonym_merge_threshold,
            data_origin=data_origin,
            synonym_generator=synonym_generator,
            include_entity_patterns=include_entity_patterns,
            exclude_entity_patterns=exclude_entity_patterns,
            curations_path=curations_path,
            global_actions=global_actions,
            label_predicate=label_predicate,
        )


class GeneOntologyParser(OntologyParser):
    _uri_regex = re.compile("^http://purl.obolibrary.org/obo/GO_[0-9]+$")

    instances: Set[str] = set()
    instances_in_dbs: Set[str] = set()

    def __init__(
        self,
        in_path: str,
        entity_class: str,
        name: str,
        query: str,
        string_scorer: Optional[StringSimilarityScorer] = None,
        synonym_merge_threshold: float = 0.70,
        data_origin: str = "unknown",
        synonym_generator: Optional[CombinatorialSynonymGenerator] = None,
        curations_path: Optional[str] = None,
        global_actions: Optional[GlobalParserActions] = None,
    ):
        super().__init__(
            in_path=in_path,
            entity_class=entity_class,
            name=name,
            string_scorer=string_scorer,
            synonym_merge_threshold=synonym_merge_threshold,
            data_origin=data_origin,
            synonym_generator=synonym_generator,
            curations_path=curations_path,
            global_actions=global_actions,
        )
        self.instances.add(name)
        self.query = query

    def populate_databases(
        self, force: bool = False, return_curations: bool = False
    ) -> Optional[List[CuratedTerm]]:
        curations = super().populate_databases(force=force, return_curations=return_curations)
        self.instances_in_dbs.add(self.name)

        if self.instances_in_dbs >= self.instances:
            # all existing instances are in the database, so we can free up
            # the memory used by the cached parsed gene ontology, which is significant.
            self.load_go.cache_clear()
        return curations

    def __del__(self):
        GeneOntologyParser.instances.discard(self.name)

    @staticmethod
    @cache
    def load_go(in_path: PathLike) -> rdflib.Graph:
        g = rdflib.Graph()
        g.parse(in_path)
        return g

    def find_kb(self, string: str) -> str:
        return self.name

    def parse_to_dataframe(self) -> pd.DataFrame:
        g = self.load_go(self.in_path)
        result = g.query(self.query)
        default_labels = []
        iris = []
        syns = []
        mapping_type = []

        # there seems to be a bug in rdflib that means the iterator sometimes exits early unless we convert to list first
        # type cast is necessary because iterating over an rdflib query result gives different types depending on the kind
        # of query, so rdflib gives a Union here, but we know it should be a ResultRow because we know we should have a
        # select query
        list_res = cast(List[rdflib.query.ResultRow], list(result))
        for row in list_res:
            idx = str(row.goid)
            label = str(row.label)
            if "obsolete" in label:
                logger.debug("skipping obsolete id: %s, %s", idx, label)
                continue
            if self._uri_regex.match(idx):
                default_labels.append(label)
                iris.append(idx)
                syns.append(str(row.synonym))
                mapping_type.append("hasExactSynonym")
        df = pd.DataFrame.from_dict(
            {DEFAULT_LABEL: default_labels, IDX: iris, SYN: syns, MAPPING_TYPE: mapping_type}
        )
        default_labels_df = df[[IDX, DEFAULT_LABEL]].drop_duplicates().copy()
        default_labels_df[SYN] = default_labels_df[DEFAULT_LABEL]
        default_labels_df[MAPPING_TYPE] = "label"

        return pd.concat([df, default_labels_df])


class BiologicalProcessGeneOntologyParser(GeneOntologyParser):
    def __init__(
        self,
        in_path: str,
        entity_class: str,
        name: str,
        string_scorer: Optional[StringSimilarityScorer] = None,
        synonym_merge_threshold: float = 0.70,
        data_origin: str = "unknown",
        synonym_generator: Optional[CombinatorialSynonymGenerator] = None,
        curations_path: Optional[str] = None,
        global_actions: Optional[GlobalParserActions] = None,
    ):
        super().__init__(
            in_path=in_path,
            entity_class=entity_class,
            string_scorer=string_scorer,
            synonym_merge_threshold=synonym_merge_threshold,
            data_origin=data_origin,
            synonym_generator=synonym_generator,
            name=name,
            curations_path=curations_path,
            global_actions=global_actions,
            query="""
                    PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
                    PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
                    PREFIX oboinowl: <http://www.geneontology.org/formats/oboInOwl#>

                    SELECT DISTINCT ?goid ?label ?synonym
                            WHERE {

                                ?goid oboinowl:hasExactSynonym ?synonym .
                                ?goid rdfs:label ?label .
                                ?goid oboinowl:hasOBONamespace "biological_process" .

                  }
            """,
        )


class MolecularFunctionGeneOntologyParser(GeneOntologyParser):
    def __init__(
        self,
        in_path: str,
        entity_class: str,
        name: str,
        string_scorer: Optional[StringSimilarityScorer] = None,
        synonym_merge_threshold: float = 0.70,
        data_origin: str = "unknown",
        synonym_generator: Optional[CombinatorialSynonymGenerator] = None,
        curations_path: Optional[str] = None,
        global_actions: Optional[GlobalParserActions] = None,
    ):
        super().__init__(
            in_path=in_path,
            entity_class=entity_class,
            string_scorer=string_scorer,
            synonym_merge_threshold=synonym_merge_threshold,
            data_origin=data_origin,
            synonym_generator=synonym_generator,
            name=name,
            curations_path=curations_path,
            global_actions=global_actions,
            query="""
                    PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
                    PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
                    PREFIX oboinowl: <http://www.geneontology.org/formats/oboInOwl#>

                    SELECT DISTINCT ?goid ?label ?synonym
                            WHERE {

                                ?goid oboinowl:hasExactSynonym ?synonym .
                                ?goid rdfs:label ?label .
                                ?goid oboinowl:hasOBONamespace "molecular_function".

                    }
            """,
        )


class CellularComponentGeneOntologyParser(GeneOntologyParser):
    def __init__(
        self,
        in_path: str,
        entity_class: str,
        name: str,
        string_scorer: Optional[StringSimilarityScorer] = None,
        synonym_merge_threshold: float = 0.70,
        data_origin: str = "unknown",
        synonym_generator: Optional[CombinatorialSynonymGenerator] = None,
        curations_path: Optional[str] = None,
        global_actions: Optional[GlobalParserActions] = None,
    ):
        super().__init__(
            in_path=in_path,
            entity_class=entity_class,
            string_scorer=string_scorer,
            synonym_merge_threshold=synonym_merge_threshold,
            data_origin=data_origin,
            synonym_generator=synonym_generator,
            name=name,
            curations_path=curations_path,
            global_actions=global_actions,
            query="""
                    PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
                    PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
                    PREFIX oboinowl: <http://www.geneontology.org/formats/oboInOwl#>

                    SELECT DISTINCT ?goid ?label ?synonym
                            WHERE {

                                ?goid oboinowl:hasExactSynonym ?synonym .
                                ?goid rdfs:label ?label .
                                ?goid oboinowl:hasOBONamespace "cellular_component" .

                    }
            """,
        )


class UberonOntologyParser(RDFGraphParser):
    """
    input should be an UBERON owl file
    e.g.
    https://www.ebi.ac.uk/ols/ontologies/uberon
    """

    def __init__(
        self,
        in_path: str,
        entity_class: str,
        name: str,
        string_scorer: Optional[StringSimilarityScorer] = None,
        synonym_merge_threshold: float = 0.70,
        data_origin: str = "unknown",
        synonym_generator: Optional[CombinatorialSynonymGenerator] = None,
        curations_path: Optional[str] = None,
        global_actions: Optional[GlobalParserActions] = None,
    ):

        super().__init__(
            in_path=in_path,
            entity_class=entity_class,
            name=name,
            uri_regex=re.compile("^http://purl.obolibrary.org/obo/UBERON_[0-9]+$"),
            synonym_predicates=(
                rdflib.URIRef("http://www.geneontology.org/formats/oboInOwl#hasExactSynonym"),
            ),
            string_scorer=string_scorer,
            synonym_merge_threshold=synonym_merge_threshold,
            data_origin=data_origin,
            synonym_generator=synonym_generator,
            curations_path=curations_path,
            global_actions=global_actions,
        )

    def find_kb(self, string: str) -> str:
        return "UBERON"


class MondoOntologyParser(OntologyParser):
    _uri_regex = re.compile("^http://purl.obolibrary.org/obo/(MONDO|HP)_[0-9]+$")
    """
    input should be a MONDO json file
    e.g.
    https://www.ebi.ac.uk/ols/ontologies/mondo
    """

    def find_kb(self, string: str):
        path = parse.urlparse(string).path
        # just the final bit, e.g. MONDO_0000123
        path_end = path.split("/")[-1]
        # we don't want the underscore or digits for the unique ID, just the ontology bit
        return path_end.split("_")[0]

    def parse_to_dataframe(self) -> pd.DataFrame:
        x = json.load(open(self.in_path, "r"))
        graph = x["graphs"][0]
        nodes = graph["nodes"]
        ids = []
        default_label_list = []
        all_syns = []
        mapping_type = []
        for i, node in enumerate(nodes):
            if not self.is_valid_iri(node["id"]):
                continue

            idx = node["id"]
            default_label = node.get("lbl")
            if default_label is None:
                # skip if no default label is available
                continue
            # add default_label to syn type
            all_syns.append(default_label)
            default_label_list.append(default_label)
            mapping_type.append("lbl")
            ids.append(idx)

            syns = node.get("meta", {}).get("synonyms", [])
            for syn_dict in syns:

                pred = syn_dict["pred"]
                if pred in {"hasExactSynonym"}:
                    mapping_type.append(pred)
                    syn = syn_dict["val"]
                    ids.append(idx)
                    default_label_list.append(default_label)
                    all_syns.append(syn)

        df = pd.DataFrame.from_dict(
            {IDX: ids, DEFAULT_LABEL: default_label_list, SYN: all_syns, MAPPING_TYPE: mapping_type}
        )
        return df

    def is_valid_iri(self, text: str) -> bool:
        match = self._uri_regex.match(text)
        return bool(match)


class EnsemblOntologyParser(OntologyParser):
    """
    input is a json from HGNC
    e.g. http://ftp.ebi.ac.uk/pub/databases/genenames/hgnc/json/hgnc_complete_set.json

    :return:
    """

    def __init__(
        self,
        in_path: str,
        entity_class: str,
        name: str,
        string_scorer: Optional[StringSimilarityScorer] = None,
        synonym_merge_threshold: float = 0.70,
        data_origin: str = "unknown",
        synonym_generator: Optional[CombinatorialSynonymGenerator] = None,
        curations_path: Optional[str] = None,
        global_actions: Optional[GlobalParserActions] = None,
    ):
        super().__init__(
            in_path=in_path,
            entity_class=entity_class,
            string_scorer=string_scorer,
            synonym_merge_threshold=synonym_merge_threshold,
            data_origin=data_origin,
            synonym_generator=synonym_generator,
            name=name,
            curations_path=curations_path,
            global_actions=global_actions,
        )

    def find_kb(self, string: str) -> str:
        return "ENSEMBL"

    def parse_to_dataframe(self) -> pd.DataFrame:

        keys_to_check = [
            "name",
            "symbol",
            "uniprot_ids",
            "alias_name",
            "alias_symbol",
            "prev_name",
            "lncipedia",
            "prev_symbol",
            "vega_id",
            "refseq_accession",
            "hgnc_id",
            "mgd_id",
            "rgd_id",
            "ccds_id",
            "pseudogene.org",
        ]

        with open(self.in_path, "r") as f:
            data = json.load(f)
        ids = []
        default_label = []
        all_syns = []
        all_mapping_type: List[str] = []
        docs = data["response"]["docs"]
        for doc in docs:

            def get_with_default_list(key: str):
                found = doc.get(key, [])
                if not isinstance(found, list):
                    found = [found]
                return found

            ensembl_gene_id = doc.get("ensembl_gene_id", None)
            name = doc.get("name", None)
            if ensembl_gene_id is None or name is None:
                continue
            else:
                # find synonyms
                synonyms: List[Tuple[str, str]] = []
                for hgnc_key in keys_to_check:
                    synonyms_this_entity = get_with_default_list(hgnc_key)
                    for potential_synonym in synonyms_this_entity:
                        synonyms.append((potential_synonym, hgnc_key))

                synonyms = list(set(synonyms))
                synonyms_strings = []
                for synonym_str, mapping_t in synonyms:
                    all_mapping_type.append(mapping_t)
                    synonyms_strings.append(synonym_str)

                num_syns = len(synonyms_strings)
                ids.extend([ensembl_gene_id] * num_syns)
                default_label.extend([name] * num_syns)
                all_syns.extend(synonyms_strings)

        df = pd.DataFrame.from_dict(
            {IDX: ids, DEFAULT_LABEL: default_label, SYN: all_syns, MAPPING_TYPE: all_mapping_type}
        )
        return df


class ChemblOntologyParser(OntologyParser):
    """
    input is a sqllite dump from Chembl, e.g.
    https://ftp.ebi.ac.uk/pub/databases/chembl/ChEMBLdb/latest/chembl_29_sqlite.tar.gz
    """

    def find_kb(self, string: str) -> str:
        return "CHEMBL"

    def parse_to_dataframe(self) -> pd.DataFrame:
        conn = sqlite3.connect(self.in_path)
        query = f"""\
            SELECT chembl_id AS {IDX}, pref_name AS {DEFAULT_LABEL}, synonyms AS {SYN}, syn_type AS {MAPPING_TYPE}
            FROM molecule_dictionary AS md
                     JOIN molecule_synonyms ms ON md.molregno = ms.molregno
            UNION ALL
            SELECT chembl_id AS {IDX}, pref_name AS {DEFAULT_LABEL}, pref_name AS {SYN}, 'pref_name' AS {MAPPING_TYPE}
            FROM molecule_dictionary
        """
        df = pd.read_sql(query, conn)
        # eliminate anything without a pref_name, as will be too big otherwise
        df = df.dropna(subset=[DEFAULT_LABEL])

        df.drop_duplicates(inplace=True)

        return df


class CLOOntologyParser(RDFGraphParser):
    """
    input is a CLO Owl file
    https://www.ebi.ac.uk/ols/ontologies/clo
    """

    def __init__(
        self,
        in_path: str,
        entity_class: str,
        name: str,
        string_scorer: Optional[StringSimilarityScorer] = None,
        synonym_merge_threshold: float = 0.70,
        data_origin: str = "unknown",
        synonym_generator: Optional[CombinatorialSynonymGenerator] = None,
        curations_path: Optional[str] = None,
        global_actions: Optional[GlobalParserActions] = None,
    ):
        super().__init__(
            in_path=in_path,
            entity_class=entity_class,
            name=name,
            uri_regex=re.compile("^http://purl.obolibrary.org/obo/CLO_[0-9]+$"),
            synonym_predicates=(
                rdflib.URIRef("http://www.geneontology.org/formats/oboInOwl#hasExactSynonym"),
            ),
            string_scorer=string_scorer,
            synonym_merge_threshold=synonym_merge_threshold,
            data_origin=data_origin,
            synonym_generator=synonym_generator,
            curations_path=curations_path,
            global_actions=global_actions,
        )

    def find_kb(self, string: str) -> str:
        return "CLO"


class CellosaurusOntologyParser(OntologyParser):
    """
    input is an obo file from cellosaurus, e.g.
    https://ftp.expasy.org/databases/cellosaurus/cellosaurus.obo
    """

    cell_line_re = re.compile("cell line", re.IGNORECASE)

    def find_kb(self, string: str) -> str:
        return "CELLOSAURUS"

    def score_and_group_ids(
        self,
        ids_and_source: IdsAndSource,
        is_symbolic: bool,
        original_syn_set: Set[str],
    ) -> Tuple[AssociatedIdSets, EquivalentIdAggregationStrategy]:
        """
        treat all synonyms as seperate cell lines

        :param ids:
        :param id_to_source:
        :param is_symbolic:
        :param original_syn_set:
        :return:
        """

        return (
            frozenset(
                EquivalentIdSet(
                    ids_and_source=frozenset((single_id_and_source,)),
                )
                for single_id_and_source in ids_and_source
            ),
            EquivalentIdAggregationStrategy.CUSTOM,
        )

    def _remove_cell_line_text(self, text: str):
        return self.cell_line_re.sub("", text).strip()

    def parse_to_dataframe(self) -> pd.DataFrame:

        ids = []
        default_labels = []
        all_syns = []
        mapping_type = []
        with open(self.in_path, "r") as f:
            id = ""
            for line in f:
                text = line.rstrip()
                if text.startswith("id:"):
                    id = text.split(" ")[1]
                elif text.startswith("name:"):
                    default_label = text[5:].strip()
                    ids.append(id)
                    # we remove "cell line" because they're all cell lines and it confuses mapping
                    default_label_no_cell_line = self._remove_cell_line_text(default_label)
                    default_labels.append(default_label_no_cell_line)
                    all_syns.append(default_label_no_cell_line)
                    mapping_type.append("name")
                # synonyms in cellosaurus are a bit of a mess, so we don't use this field for now. Leaving this here
                # in case they improve at some point
                # elif text.startswith("synonym:"):
                #     match = self._synonym_regex.match(text)
                #     if match is None:
                #         raise ValueError(
                #             """synonym line does not match our synonym regex.
                #             Either something is wrong with the file, or it has updated
                #             and our regex is not correct/general enough."""
                #         )
                #     ids.append(id)
                #     default_labels.append(default_label)
                #
                #     all_syns.append(self._remove_cell_line_text(match.group("syn")))
                #     mapping_type.append(match.group("mapping"))
                else:
                    pass
        df = pd.DataFrame.from_dict(
            {IDX: ids, DEFAULT_LABEL: default_labels, SYN: all_syns, MAPPING_TYPE: mapping_type}
        )
        return df

    _synonym_regex = re.compile(
        r"""^synonym:      # line that begins synonyms
        \s*                # any amount of whitespace (standardly a single space)
        "(?P<syn>[^"]*)"   # a quoted string - capture this as a named match group 'syn'
        \s*                # any amount of separating whitespace (standardly a single space)
        (?P<mapping>\w*)   # a sequence of word characters representing the mapping type
        \s*                # any amount of separating whitespace (standardly a single space)
        \[\]               # an open and close bracket at the end of the string
        $""",
        re.VERBOSE,
    )


class MeddraOntologyParser(OntologyParser):
    """
    input is an unzipped directory to a Meddra release (Note, requires licence). This
    should contain the files 'mdhier.asc' and 'llt.asc'.
    """

    def __init__(
        self,
        in_path: str,
        entity_class: str,
        name: str,
        string_scorer: Optional[StringSimilarityScorer] = None,
        synonym_merge_threshold: float = 0.70,
        data_origin: str = "unknown",
        synonym_generator: Optional[CombinatorialSynonymGenerator] = None,
        curations_path: Optional[str] = None,
        global_actions: Optional[GlobalParserActions] = None,
        exclude_socs: Iterable[str] = (
            "Surgical and medical procedures",
            "Social circumstances",
            "Investigations",
        ),
    ):
        super().__init__(
            in_path=in_path,
            entity_class=entity_class,
            string_scorer=string_scorer,
            synonym_merge_threshold=synonym_merge_threshold,
            data_origin=data_origin,
            synonym_generator=synonym_generator,
            curations_path=curations_path,
            global_actions=global_actions,
            name=name,
        )

        self.exclude_socs = exclude_socs

    _mdhier_asc_col_names = (
        "pt_code",
        "hlt_code",
        "hlgt_code",
        "soc_code",
        "pt_name",
        "hlt_name",
        "hlgt_name",
        "soc_name",
        "soc_abbrev",
        "null_field",
        "pt_soc_code",
        "primary_soc_fg",
        "NULL",
    )

    _llt_asc_column_names = (
        "llt_code",
        "llt_name",
        "pt_code",
        "llt_whoart_code",
        "llt_harts_code",
        "llt_costart_sym",
        "llt_icd9_code",
        "llt_icd9cm_code",
        "llt_icd10_code",
        "llt_currency",
        "llt_jart_code",
        "NULL",
    )

    def find_kb(self, string: str) -> str:
        return "MEDDRA"

    def parse_to_dataframe(self) -> pd.DataFrame:
        # hierarchy path
        mdheir_path = os.path.join(self.in_path, "mdhier.asc")
        # low level term path
        llt_path = os.path.join(self.in_path, "llt.asc")
        hier_df = pd.read_csv(
            mdheir_path,
            sep="$",
            header=None,
            names=self._mdhier_asc_col_names,
            dtype="string",
        )
        hier_df = hier_df[~hier_df["soc_name"].isin(self.exclude_socs)]

        llt_df = pd.read_csv(
            llt_path,
            sep="$",
            header=None,
            names=self._llt_asc_column_names,
            usecols=("llt_name", "pt_code"),
            dtype="string",
        )
        llt_df = llt_df.dropna(axis=1)

        ids = []
        default_labels = []
        all_syns = []
        mapping_type = []
        soc_names = []
        soc_codes = []

        for i, row in hier_df.iterrows():
            idx = row["pt_code"]
            pt_name = row["pt_name"]
            soc_name = row["soc_name"]
            soc_code = row["soc_code"]
            llts = llt_df[llt_df["pt_code"] == idx]
            ids.append(idx)
            default_labels.append(pt_name)
            all_syns.append(pt_name)
            soc_names.append(soc_name)
            soc_codes.append(soc_code)
            mapping_type.append("meddra_link")
            for j, llt_row in llts.iterrows():
                ids.append(idx)
                default_labels.append(pt_name)
                soc_names.append(soc_name)
                soc_codes.append(soc_code)
                all_syns.append(llt_row["llt_name"])
                mapping_type.append("meddra_link")

        for i, row in (
            hier_df[["hlt_code", "hlt_name", "soc_name", "soc_code"]].drop_duplicates().iterrows()
        ):
            ids.append(row["hlt_code"])
            default_labels.append(row["hlt_name"])
            soc_names.append(row["soc_name"])
            soc_codes.append(row["soc_code"])
            all_syns.append(row["hlt_name"])
            mapping_type.append("meddra_link")
        for i, row in (
            hier_df[["hlgt_code", "hlgt_name", "soc_name", "soc_code"]].drop_duplicates().iterrows()
        ):
            ids.append(row["hlgt_code"])
            default_labels.append(row["hlgt_name"])
            soc_names.append(row["soc_name"])
            soc_codes.append(row["soc_code"])
            all_syns.append(row["hlgt_name"])
            mapping_type.append("meddra_link")
        df = pd.DataFrame.from_dict(
            {
                IDX: ids,
                DEFAULT_LABEL: default_labels,
                SYN: all_syns,
                MAPPING_TYPE: mapping_type,
                "soc_name": soc_names,
                "soc_code": soc_codes,
            }
        )
        return df


class CLOntologyParser(RDFGraphParser):
    """
    input should be an CL owl file
    e.g.
    https://www.ebi.ac.uk/ols/ontologies/cl
    """

    def __init__(
        self,
        in_path: str,
        entity_class: str,
        name: str,
        string_scorer: Optional[StringSimilarityScorer] = None,
        synonym_merge_threshold: float = 0.7,
        data_origin: str = "unknown",
        synonym_generator: Optional[CombinatorialSynonymGenerator] = None,
        include_entity_patterns: Optional[Iterable[PredicateAndValue]] = None,
        exclude_entity_patterns: Optional[Iterable[PredicateAndValue]] = None,
        curations_path: Optional[str] = None,
        global_actions: Optional[GlobalParserActions] = None,
    ):

        super().__init__(
            in_path=in_path,
            entity_class=entity_class,
            name=name,
            uri_regex=re.compile("^http://purl.obolibrary.org/obo/CL_[0-9]+$"),
            synonym_predicates=(
                rdflib.URIRef("http://www.geneontology.org/formats/oboInOwl#hasExactSynonym"),
            ),
            string_scorer=string_scorer,
            synonym_merge_threshold=synonym_merge_threshold,
            data_origin=data_origin,
            synonym_generator=synonym_generator,
            include_entity_patterns=include_entity_patterns,
            exclude_entity_patterns=exclude_entity_patterns,
            curations_path=curations_path,
            global_actions=global_actions,
        )

    def find_kb(self, string: str) -> str:
        return "CL"


class HGNCGeneFamilyParser(OntologyParser):

    syn_column_keys = {"Family alias", "Common root gene symbol"}

    def find_kb(self, string: str) -> str:
        return "HGNC_GENE_FAMILY"

    def parse_to_dataframe(self) -> pd.DataFrame:
        df = pd.read_csv(self.in_path, sep="\t")
        data = []
        for family_id, row in (
            df.groupby(by="Family ID").agg(lambda col_series: set(col_series.dropna())).iterrows()
        ):
            # in theory, there should only be one family name per ID
            assert len(row["Family name"]) == 1
            default_label = next(iter(row["Family name"]))
            data.append(
                {
                    SYN: default_label,
                    MAPPING_TYPE: "Family name",
                    DEFAULT_LABEL: default_label,
                    IDX: family_id,
                }
            )
            data.extend(
                {
                    SYN: syn,
                    MAPPING_TYPE: key,
                    DEFAULT_LABEL: default_label,
                    IDX: family_id,
                }
                for key in self.syn_column_keys
                for syn in row[key]
            )
        return pd.DataFrame.from_records(data)


class TabularOntologyParser(OntologyParser):
    """
    For already tabulated data.
    """

    def __init__(
        self,
        in_path: str,
        entity_class: str,
        name: str,
        string_scorer: Optional[StringSimilarityScorer] = None,
        synonym_merge_threshold: float = 0.7,
        data_origin: str = "unknown",
        synonym_generator: Optional[CombinatorialSynonymGenerator] = None,
        curations_path: Optional[str] = None,
        global_actions: Optional[GlobalParserActions] = None,
        **kwargs,
    ):
        """

        :param in_path:
        :param entity_class:
        :param name:
        :param string_scorer:
        :param synonym_merge_threshold:
        :param data_origin:
        :param synonym_generator:
        :param curations_path:
        :param kwargs: passed to pandas.read_csv
        """
        super().__init__(
            in_path=in_path,
            entity_class=entity_class,
            name=name,
            string_scorer=string_scorer,
            synonym_merge_threshold=synonym_merge_threshold,
            data_origin=data_origin,
            synonym_generator=synonym_generator,
            curations_path=curations_path,
            global_actions=global_actions,
        )
        self._raw_dataframe = pd.read_csv(self.in_path, **kwargs)

    def parse_to_dataframe(self) -> pd.DataFrame:
        """
        Assume input file is already in correct format.

        Inherit and override this method if different behaviour is required.

        :return:
        """
        return self._raw_dataframe

    def find_kb(self, string: str) -> str:
        return self.name


class ATCDrugClassificationParser(TabularOntologyParser):
    """
    Parser for the ATC Drug classification dataset.

    This requires a licence from WHO, available at
    https://www.who.int/tools/atc-ddd-toolkit/atc-classification .

    """

    def __init__(
        self,
        in_path: str,
        entity_class: str,
        name: str,
        string_scorer: Optional[StringSimilarityScorer] = None,
        synonym_merge_threshold: float = 0.70,
        data_origin: str = "unknown",
        synonym_generator: Optional[CombinatorialSynonymGenerator] = None,
        curations_path: Optional[str] = None,
        global_actions: Optional[GlobalParserActions] = None,
    ):
        super().__init__(
            in_path=in_path,
            entity_class=entity_class,
            name=name,
            string_scorer=string_scorer,
            synonym_merge_threshold=synonym_merge_threshold,
            data_origin=data_origin,
            synonym_generator=synonym_generator,
            curations_path=curations_path,
            global_actions=global_actions,
            sep="     ",
            header=None,
            names=["code", "level_and_description"],
            # Because the c engine can't handle multi-char sep
            # removing this results in the same behaviour, but
            # pandas logs a warning.
            engine="python",
        )

    levels_to_ignore = {"1", "2", "3"}

    def parse_to_dataframe(self) -> pd.DataFrame:
        # for some reason, the level and description codes are merged, so we need to fix this here
        df = self._raw_dataframe.applymap(str.strip)
        res_df = pd.DataFrame()
        res_df[[MAPPING_TYPE, DEFAULT_LABEL]] = df.apply(
            lambda row: [row["level_and_description"][0], row["level_and_description"][1:]],
            axis=1,
            result_type="expand",
        )
        res_df[IDX] = df["code"]
        res_df = res_df[~res_df[MAPPING_TYPE].isin(self.levels_to_ignore)]
        res_df[SYN] = res_df[DEFAULT_LABEL]
        return res_df


class StatoParser(RDFGraphParser):
    """
    Parse stato: input should be an owl file.

    Available at e.g.
    https://www.ebi.ac.uk/ols/ontologies/stato .
    """

    def __init__(
        self,
        in_path: str,
        entity_class: str,
        name: str,
        string_scorer: Optional[StringSimilarityScorer] = None,
        synonym_merge_threshold: float = 0.7,
        data_origin: str = "unknown",
        synonym_generator: Optional[CombinatorialSynonymGenerator] = None,
        include_entity_patterns: Optional[Iterable[PredicateAndValue]] = None,
        exclude_entity_patterns: Optional[Iterable[PredicateAndValue]] = None,
        curations_path: Optional[str] = None,
        global_actions: Optional[GlobalParserActions] = None,
    ):

        super().__init__(
            in_path=in_path,
            entity_class=entity_class,
            name=name,
            uri_regex=re.compile("^http://purl.obolibrary.org/obo/(OBI|STATO)_[0-9]+$"),
            synonym_predicates=(rdflib.URIRef("http://purl.obolibrary.org/obo/IAO_0000111"),),
            string_scorer=string_scorer,
            synonym_merge_threshold=synonym_merge_threshold,
            data_origin=data_origin,
            synonym_generator=synonym_generator,
            include_entity_patterns=include_entity_patterns,
            exclude_entity_patterns=exclude_entity_patterns,
            curations_path=curations_path,
            global_actions=global_actions,
        )

    def find_kb(self, string: str) -> str:
        return "OBI" if "OBI" in string else "STATO"



import dataclasses
from typing import List, Protocol, Callable, Iterable, Optional
import urllib

from kazu.data.data import (
    Document,
    Entity,
    Mapping,
    StringMatchConfidence,
    DisambiguationConfidence,
    MentionConfidence,
)
from kazu.steps import Step, document_iterating_step

EntityFilterFn = Callable[[Entity], bool]
MappingFilterFn = Callable[[Mapping], bool]


class CleanupAction(Protocol):
    def cleanup(self, doc: Document):
        raise NotImplementedError


class MappingFilterCleanupAction:
    def __init__(self, filter_fns: List[MappingFilterFn]):
        self.filter_fns = filter_fns

    def cleanup(self, doc: Document):
        for entity in doc.get_entities():
            entity.mappings = {
                mapping
                for mapping in entity.mappings
                if not any(f(mapping) for f in self.filter_fns)
            }


class EntityFilterCleanupAction:
    def __init__(self, filter_fns: List[EntityFilterFn]):
        self.filter_fns = filter_fns

    def cleanup(self, doc: Document):
        for section in doc.sections:
            section.entities = [
                entity for entity in section.entities if not any(f(entity) for f in self.filter_fns)
            ]


class DropMappingsByConfidenceMappingFilter:
    def __init__(
        self,
        string_match_ranks_to_drop: Iterable[StringMatchConfidence],
        disambiguation_ranks_to_drop: Iterable[DisambiguationConfidence],
    ):
        self.string_match_ranks_to_drop = set(string_match_ranks_to_drop)
        self.disambiguation_ranks_to_drop = set(disambiguation_ranks_to_drop)

    def __call__(self, mapping: Mapping) -> bool:

        return (
            mapping.string_match_confidence in self.string_match_ranks_to_drop
            or mapping.disambiguation_confidence in self.disambiguation_ranks_to_drop
        )


class DropUnmappedEntityFilter:
    def __init__(
        self,
        from_ent_namespaces: Optional[Iterable[str]] = None,
        min_confidence_level: Optional[MentionConfidence] = MentionConfidence.PROBABLE,
    ):
        self.min_confidence_level = min_confidence_level
        self.from_ent_namespaces = None if from_ent_namespaces is None else set(from_ent_namespaces)

    def __call__(self, ent: Entity) -> bool:
        relevant_namespace = (
            self.from_ent_namespaces is None or ent.namespace in self.from_ent_namespaces
        )

        if self.min_confidence_level is None:
            return relevant_namespace and len(ent.mappings) == 0
        else:
            return (
                relevant_namespace
                and len(ent.mappings) == 0
                and ent.mention_confidence < self.min_confidence_level
            )


class StripMappingURIsAction:
    """Strip the IDs in :class:`kazu.data.data.Mapping` to just the final part of the URI.

    For example, this will turn http://purl.obolibrary.org/obo/MONDO_0004979 into just MONDO_004979.

    If you don't want URI stripping at all, don't use this Action as part of the CleanupStep/in the pipeline.
    """

    def __init__(self, parsers_to_strip: Optional[Iterable[str]] = None):
        """

        :param parsers_to_strip: if you only want to strip URIs for some parsers and not others,
            provide the parsers to strip here. Otherwise, all parsers will have their IDs stripped.
            This prevents having to keep the full list of parsers in sync here.
        """
        self.parsers_to_strip = parsers_to_strip

    @staticmethod
    def _strip_uri(idx):
        url = urllib.parse.urlparse(idx)
        if url.scheme == "":
            # not a url
            new_idx = idx
        else:
            new_idx = url.path.split("/")[-1]
        return new_idx

    def cleanup(self, doc: Document):
        for entity in doc.get_entities():
            new_mappings = set()
            for mapping in entity.mappings:
                if (
                    self.parsers_to_strip is not None
                    and mapping.parser_name not in self.parsers_to_strip
                ):
                    new_mappings.add(mapping)
                else:
                    new_mappings.add(dataclasses.replace(mapping, idx=self._strip_uri(mapping.idx)))
            entity.mappings = new_mappings


class CleanupStep(Step):
    def __init__(self, cleanup_actions: List[CleanupAction]):
        self.cleanup_actions = cleanup_actions

    @document_iterating_step
    def __call__(self, doc: Document) -> None:
        for cleanup_action in self.cleanup_actions:
            cleanup_action.cleanup(doc)
